#    builder.py -- Classes for building packages
#    Copyright (C) 2006, 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import shutil
import subprocess
import os

from bzrlib.trace import note

from bzrlib.plugins.builddeb.errors import (
                    NoSourceDirError,
                    BuildFailedError,
                    )
from bzrlib.plugins.builddeb.quilt import quilt_push_all
from bzrlib.plugins.builddeb.util import (
    get_parent_dir,
    subprocess_setup,
    FORMAT_3_0_QUILT,
    )


class DebBuild(object):
    """The object that does the building work."""

    def __init__(self, distiller, target_dir, builder, use_existing=False):
        """Create a builder.

        :param distiller: the SourceDistiller that will get the source to
            build.
        :param target_dir: the directory in which to do all the work.
        :param builder: the build command to use.
        :param use_existing: whether to re-use the target_dir if it exists.
        """
        self.distiller = distiller
        self.target_dir = target_dir
        self.builder = builder
        self.use_existing = use_existing

    def prepare(self):
        """Do any preparatory steps that should be run before the build.

        It checks that everything is well, and that some needed dirs are
        created.
        """
        parent_dir = get_parent_dir(self.target_dir)
        if parent_dir != '' and not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        if os.path.exists(self.target_dir):
            if not self.use_existing:
                note("Purging the build dir: %s", self.target_dir)
                shutil.rmtree(self.target_dir)
            else:
                note("Not purging build dir as requested: %s",
                        self.target_dir)
        else:
            if self.use_existing:
                raise NoSourceDirError

    def export(self, apply_quilt_patches=True):
        self.distiller.distill(self.target_dir)
        if apply_quilt_patches:
            self._apply_quilt_patches()

    def _apply_quilt_patches(self):
        if not os.path.isfile(os.path.join(self.target_dir, "debian/patches/series")):
            return
        format_path = os.path.join(self.target_dir, "debian/source/format")
        if not os.path.isfile(format_path):
            return
        with file(format_path, 'r') as f:
            if f.read().strip() == FORMAT_3_0_QUILT:
                quilt_push_all(os.path.abspath(self.target_dir))

    def build(self):
        """This builds the package using the supplied command."""
        note("Building the package in %s, using %s", self.target_dir,
                self.builder)
        proc = subprocess.Popen(self.builder, shell=True, cwd=self.target_dir,
                preexec_fn=subprocess_setup)
        proc.wait()
        if proc.returncode != 0:
            raise BuildFailedError

    def clean(self):
        """This removes the build directory."""
        note("Cleaning build dir: %s", self.target_dir)
        shutil.rmtree(self.target_dir)



#    test_hooks.py -- Tests for builddeb hooks.
#    Copyright (C) 2006 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os

from bzrlib.plugins.builddeb.config import DebBuildConfig
from bzrlib.plugins.builddeb.errors import HookFailedError
from bzrlib.plugins.builddeb.hooks import run_hook
from bzrlib.plugins.builddeb.tests import TestCaseInTempDir

class MockTree:
  def abspath(self, relpath):
    return os.path.abspath(relpath)

class HookTests(TestCaseInTempDir):

  default_conf = 'default.conf'
  local_conf = 'local.conf'

  def test_run_hook_allows_no_hook_defined(self):
    f = open(self.default_conf, 'wb')
    f.close()
    config = DebBuildConfig([(self.default_conf, False)])
    run_hook(MockTree(), 'pre-build', config)

  def test_run_hook_raises_when_hook_fails(self):
    f = open(self.default_conf, 'wb')
    try:
      f.write('[HOOKS]\npre-build = false\n')
    finally:
      f.close()
    config = DebBuildConfig([(self.default_conf, False)])
    self.assertRaises(HookFailedError, run_hook, MockTree(), 'pre-build', config)

  def test_run_hook_when_hook_passes(self):
    f = open(self.default_conf, 'wb')
    try:
      f.write('[HOOKS]\npre-build = true\n')
    finally:
      f.close()
    config = DebBuildConfig([(self.default_conf, False)])
    run_hook(MockTree(), 'pre-build', config)

  def test_run_hook_uses_cwd_by_default(self):
    f = open(self.default_conf, 'wb')
    try:
      f.write('[HOOKS]\npre-build = touch a\n')
    finally:
      f.close()
    config = DebBuildConfig([(self.default_conf, False)])
    run_hook(MockTree(), 'pre-build', config)
    self.assertPathExists('a')

  def test_run_hook_uses_passed_wd(self):
    os.mkdir('dir')
    f = open(self.default_conf, 'wb')
    try:
      f.write('[HOOKS]\npre-build = touch a\n')
    finally:
      f.close()
    config = DebBuildConfig([(self.default_conf, False)])
    run_hook(MockTree(), 'pre-build', config, wd='dir')
    self.assertPathExists('dir/a')

  def test_run_hook_uses_shell(self):
    f = open(self.default_conf, 'wb')
    try:
      f.write('[HOOKS]\npost-build = touch a && touch b\n')
    finally:
      f.close()
    config = DebBuildConfig([(self.default_conf, False)])
    run_hook(MockTree(), 'post-build', config)
    self.assertPathExists('a')
    self.assertPathExists('b')

  def test_run_hook_uses_local_over_global(self):
    f = open(self.default_conf, 'wb')
    try:
      f.write('[HOOKS]\npost-build = touch a\n')
    finally:
      f.close()
    f = open(self.local_conf, 'wb')
    try:
      f.write('[HOOKS]\npost-build = touch b\n')
    finally:
      f.close()
    config = DebBuildConfig([(self.local_conf, False),
                             (self.default_conf, False)])
    run_hook(MockTree(), 'post-build', config)
    self.assertPathDoesNotExist('a')
    self.assertPathExists('b')

# vim: ts=2 sts=2 sw=2



# This file is a modified copy of bzrtools' upstream_import.py, last changed in
# bzrtools 1.14.0.

"""Import upstream source into a branch"""

import errno
import os
from StringIO import StringIO
import stat
import tarfile
import zipfile

from bzrlib import generate_ids
from bzrlib.bzrdir import BzrDir
from bzrlib.errors import NoSuchFile, BzrCommandError, NotBranchError
from bzrlib.osutils import (pathjoin, isdir, file_iterator, basename,
                            file_kind, splitpath, normpath, walkdirs)
from bzrlib.trace import warning
from bzrlib.transform import TreeTransform, resolve_conflicts, cook_conflicts
from bzrlib.workingtree import WorkingTree
from bzrlib.plugins.builddeb.bzrtools_bzrtools import open_from_url
from bzrlib.plugins.builddeb.errors import UnknownType

class ZipFileWrapper(object):

    def __init__(self, fileobj, mode):
        self.zipfile = zipfile.ZipFile(fileobj, mode)

    def getmembers(self):
        for info in self.zipfile.infolist():
            yield ZipInfoWrapper(self.zipfile, info)

    def extractfile(self, infowrapper):
        return StringIO(self.zipfile.read(infowrapper.name))

    def add(self, filename):
        if isdir(filename):
            self.zipfile.writestr(filename+'/', '')
        else:
            self.zipfile.write(filename)

    def close(self):
        self.zipfile.close()


class ZipInfoWrapper(object):

    def __init__(self, zipfile, info):
        self.info = info
        self.type = None
        self.name = info.filename
        self.zipfile = zipfile
        self.mode = 0666

    def isdir(self):
        # Really? Eeeew!
        return bool(self.name.endswith('/'))

    def isreg(self):
        # Really? Eeeew!
        return not self.isdir()


files_to_ignore = set(
    ['.shelf', '.bzr', '.bzr.backup', '.bzrtags',
     '.bzr-builddeb'])


class DirWrapper(object):
    def __init__(self, fileobj, mode='r'):
        assert mode == 'r', mode
        self.root = os.path.realpath(fileobj.read())

    def __repr__(self):
        return 'DirWrapper(%r)' % self.root

    def getmembers(self):
        for _, dir_block in walkdirs(self.root):
            for relpath, _, _, stat_result, _ in dir_block:
                yield FileInfo(self.root, relpath, stat_result)

    def extractfile(self, member):
        return open(member.fullpath)


class FileInfo(object):

    def __init__(self, root, filepath, stat):
        self.fullpath = pathjoin(root, filepath)
        self.root = root
        if filepath != '':
            self.name = pathjoin(basename(root), filepath)
        else:
            self.name = basename(root)
        self.type = None
        self.mode = stat.st_mode
        if self.isdir():
            self.name += '/'

    def __repr__(self):
        return 'FileInfo(%r)' % self.name

    def isreg(self):
        return stat.S_ISREG(self.mode)

    def isdir(self):
        return stat.S_ISDIR(self.mode)

    def issym(self):
        if stat.S_ISLNK(self.mode):
            self.linkname = os.readlink(self.fullpath)
            return True
        else:
            return False


def top_path(path):
    """Return the top directory given in a path."""
    components = splitpath(normpath(path))
    if len(components) > 0:
        return components[0]
    else:
        return ''


def common_directory(names):
    """Determine a single directory prefix from a list of names"""
    prefixes = set()
    prefixes.update(map(top_path, names))
    if '' in prefixes:
        prefixes.remove('')
    if len(prefixes) != 1:
        return None
    prefix = prefixes.pop()
    if prefix == '':
        return None
    return prefix


def do_directory(tt, trans_id, tree, relative_path, path):
    if isdir(path) and tree.path2id(relative_path) is not None:
        tt.cancel_deletion(trans_id)
    else:
        tt.create_directory(trans_id)


def add_implied_parents(implied_parents, path):
    """Update the set of implied parents from a path"""
    parent = os.path.dirname(path)
    if parent in implied_parents:
        return
    implied_parents.add(parent)
    add_implied_parents(implied_parents, parent)


def names_of_files(tar_file):
    for member in tar_file.getmembers():
        if member.type != "g":
            yield member.name


def should_ignore(relative_path):
    parts = splitpath(relative_path)
    if not parts:
        return False
    for part in parts:
        if part in files_to_ignore:
            return True
        if part.endswith(',v'):
            return True


def import_tar(tree, tar_input, file_ids_from=None, target_tree=None):
    """Replace the contents of a working directory with tarfile contents.
    The tarfile may be a gzipped stream.  File ids will be updated.
    """
    tar_file = tarfile.open('lala', 'r', tar_input)
    import_archive(tree, tar_file, file_ids_from=file_ids_from,
            target_tree=target_tree)

def import_zip(tree, zip_input, file_ids_from=None, target_tree=None):
    zip_file = ZipFileWrapper(zip_input, 'r')
    import_archive(tree, zip_file, file_ids_from=file_ids_from,
            target_tree=target_tree)

def import_dir(tree, dir, file_ids_from=None, target_tree=None):
    dir_input = StringIO(dir)
    dir_file = DirWrapper(dir_input)
    import_archive(tree, dir_file, file_ids_from=file_ids_from,
            target_tree=target_tree)

def import_archive(tree, archive_file, file_ids_from=None, target_tree=None):
    if file_ids_from is None:
        file_ids_from = []
    for other_tree in file_ids_from:
        other_tree.lock_read()
    try:
        return _import_archive(tree, archive_file, file_ids_from,
                target_tree=target_tree)
    finally:
        for other_tree in file_ids_from:
            other_tree.unlock()

def _get_paths_to_process(archive_file, prefix, implied_parents):
    to_process = set()
    for member in archive_file.getmembers():
        if member.type == 'g':
            # type 'g' is a header
            continue
        relative_path = member.name
        relative_path = normpath(relative_path)
        relative_path = relative_path.lstrip('/')
        if prefix is not None:
            relative_path = relative_path[len(prefix)+1:]
            relative_path = relative_path.rstrip('/')
        if relative_path == '' or relative_path == '.':
            continue
        if should_ignore(relative_path):
            continue
        add_implied_parents(implied_parents, relative_path)
        to_process.add((relative_path, member))
    return to_process


def _import_archive(tree, archive_file, file_ids_from, target_tree=None):
    prefix = common_directory(names_of_files(archive_file))
    tt = TreeTransform(tree)
    try:
        removed = set()
        for path, entry in tree.iter_entries_by_dir():
            if entry.parent_id is None:
                continue
            trans_id = tt.trans_id_tree_path(path)
            tt.delete_contents(trans_id)
            removed.add(path)

        added = set()
        implied_parents = set()
        seen = set()
        to_process = _get_paths_to_process(archive_file, prefix,
                implied_parents)
        renames = {}

        # First we find the renames
        other_trees = file_ids_from[:]
        if target_tree is not None:
            other_trees.insert(0, target_tree)
        for other_tree in other_trees:
            for relative_path, member in to_process:
                trans_id = tt.trans_id_tree_path(relative_path)
                existing_file_id = tt.tree_file_id(trans_id)
                target_id = other_tree.path2id(relative_path)
                if (target_id is not None
                    and target_id != existing_file_id
                    and target_id not in renames):
                    renames[target_id] = relative_path

        # The we do the work
        for relative_path, member in to_process:
            trans_id = tt.trans_id_tree_path(relative_path)
            added.add(relative_path.rstrip('/'))
            # To handle renames, we need to not use the preserved file id, rather
            # we need to lookup the file id in target_tree, if there is one. If
            # there isn't, we should use the one in the current tree, and failing
            # that we will allocate one. In this importer we want the
            # target_tree to be authoritative about id2path, which is why we
            # consult it first.
            existing_file_id = tt.tree_file_id(trans_id)
            # If we find an id that we know we are going to assign to
            # different path as it has been renamed in one of the
            # file_ids_from trees then we ignore the one in this tree.
            if existing_file_id in renames:
                if relative_path != renames[existing_file_id]:
                    existing_file_id = None
            found_file_id = None
            if target_tree is not None:
                found_file_id = target_tree.path2id(relative_path)
                if found_file_id in renames:
                    if renames[found_file_id] != relative_path:
                        found_file_id = None
            if found_file_id is None and existing_file_id is None:
                for other_tree in file_ids_from:
                    found_file_id = other_tree.path2id(relative_path)
                    if found_file_id is not None:
                        if found_file_id in renames:
                            if renames[found_file_id] != relative_path:
                                found_file_id = None
                                continue
                        break
            if (found_file_id is not None
                and found_file_id != existing_file_id):
                # Found a specific file id in one of the source trees
                tt.version_file(found_file_id, trans_id)
                if existing_file_id is not None:
                    # We need to remove the existing file so it can be
                    # replaced by the file (and file id) from the
                    # file_ids_from tree.
                    tt.delete_versioned(trans_id)
                trans_id = tt.trans_id_file_id(found_file_id)

            if not found_file_id and not existing_file_id:
                # No file_id in any of the source trees and no file id in the base
                # tree.
                name = basename(member.name.rstrip('/'))
                file_id = generate_ids.gen_file_id(name)
                tt.version_file(file_id, trans_id)
            path = tree.abspath(relative_path)
            if member.name in seen:
                if tt.final_kind(trans_id) == 'file':
                    tt.set_executability(None, trans_id)
                tt.cancel_creation(trans_id)
            seen.add(member.name)
            if member.isreg():
                tt.create_file(file_iterator(archive_file.extractfile(member)),
                               trans_id)
                executable = (member.mode & 0111) != 0
                tt.set_executability(executable, trans_id)
            elif member.isdir():
                do_directory(tt, trans_id, tree, relative_path, path)
            elif member.issym():
                tt.create_symlink(member.linkname, trans_id)
            else:
                raise UnknownType(relative_path)

        for relative_path in implied_parents.difference(added):
            if relative_path == "":
                continue
            trans_id = tt.trans_id_tree_path(relative_path)
            path = tree.abspath(relative_path)
            do_directory(tt, trans_id, tree, relative_path, path)
            if tt.tree_file_id(trans_id) is None:
                found = False
                for other_tree in file_ids_from:
                    other_tree.lock_read()
                    try:
                        if other_tree.has_filename(relative_path):
                            file_id = other_tree.path2id(relative_path)
                            if file_id is not None:
                                tt.version_file(file_id, trans_id)
                                found = True
                                break
                    finally:
                        other_tree.unlock()
                if not found:
                    # Should this really use the trans_id as the
                    # file_id?
                    tt.version_file(trans_id, trans_id)
            added.add(relative_path)

        for path in removed.difference(added):
            tt.unversion_file(tt.trans_id_tree_path(path))

        for conflict in cook_conflicts(resolve_conflicts(tt), tt):
            warning(conflict)
        tt.apply()
    finally:
        tt.finalize()


def do_import(source, tree_directory=None):
    """Implementation of import command.  Intended for UI only"""
    if tree_directory is not None:
        try:
            tree = WorkingTree.open(tree_directory)
        except NotBranchError:
            if not os.path.exists(tree_directory):
                os.mkdir(tree_directory)
            branch = BzrDir.create_branch_convenience(tree_directory)
            tree = branch.bzrdir.open_workingtree()
    else:
        tree = WorkingTree.open_containing('.')[0]
    tree.lock_write()
    try:
        if tree.changes_from(tree.basis_tree()).has_changed():
            raise BzrCommandError("Working tree has uncommitted changes.")

        if (source.endswith('.tar') or source.endswith('.tar.gz') or
            source.endswith('.tar.bz2')) or source.endswith('.tgz'):
            try:
                tar_input = open_from_url(source)
                if source.endswith('.bz2'):
                    tar_input = StringIO(tar_input.read().decode('bz2'))
            except IOError, e:
                if e.errno == errno.ENOENT:
                    raise NoSuchFile(source)
            try:
                import_tar(tree, tar_input)
            finally:
                tar_input.close()
        elif source.endswith('.zip'):
            import_zip(tree, open_from_url(source))
        elif file_kind(source) == 'directory':
            s = StringIO(source)
            s.seek(0)
            import_dir(tree, s)
        else:
            raise BzrCommandError('Unhandled import source')
    finally:
        tree.unlock()



#    test_builddeb.py -- Blackbox tests for builddeb.
#    Copyright (C) 2009 Canonical Ltd.
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from bzrlib.merge import Merger
import os
import string

from bzrlib import (
    errors,
    version_info as bzrlib_version,
    )
from bzrlib.tests import TestNotApplicable
from bzrlib.plugins.builddeb import pre_merge_fix_ancestry
from bzrlib.plugins.builddeb.tests import BuilddebTestCase


_Debian_changelog = '''\
ipsec-tools (%s) unstable; urgency=high

  * debian packaging -- %s

 -- Nico Golde <nion@debian.org>  Tue, %02d May 2009 13:26:14 +0200

'''

_Ubuntu_changelog = '''\
ipsec-tools (%s) karmic; urgency=low

  * ubuntu packaging -- %s

 -- Jamie Strandboge <jamie@ubuntu.com>  Fri, %02d Jul 2009 13:24:17 -0500

'''


def _prepend_log(text, path):
    content = open(path).read()
    fh = open(path, 'wb')
    try:
        fh.write(text+content)
    finally:
        fh.close()


class TestMergePackageBB(BuilddebTestCase):

    def test_merge_package_shared_rev_conflict(self):
        """Source upstream conflicts with target packaging -> Error.

        The debian upstream and the ubuntu packaging branches will differ
        with respect to the content of the file 'c'.

        The conflict cannot be resolved by fix_ancestry_as_needed().
        The `SharedUpstreamConflictsWithTargetPackaging` exception is
        thrown instead.
        """
        target, _source = self.make_conflicting_branches_setup()
        os.chdir('ubup-o')
        merge_source = '../debp-n'
        self.run_bzr_error(
            ['branches for the merge source and target have diverged'],
            'merge-package %s' % merge_source)

    def test_pre_merge_hook_shared_rev_conflict(self):
        """Source upstream conflicts with target packaging -> Error.

        The debian upstream and the ubuntu packaging branches will differ
        with respect to the content of the file 'c'.

        The conflict cannot be resolved by fix_ancestry_as_needed().
        The `SharedUpstreamConflictsWithTargetPackaging` exception is
        thrown instead.
        """
        target, _source = self.make_conflicting_branches_setup()
        os.chdir('ubup-o')
        merge_source = '../debp-n'
        try:
            Merger.hooks.install_named_hook(
                "pre_merge", pre_merge_fix_ancestry, "fix ancestry")
        except errors.UnknownHook:
            raise TestNotApplicable("pre_merge hook requires bzr 2.5")
        self.run_bzr_error(
            ['branches for the merge source and target have diverged'],
            'merge %s' % merge_source)

    def make_conflicting_branches_setup(self):
        """
        Set up the following test configuration (debian upstream newer).

        debian-upstream                 ,------------------H
                           A-----------B                    \
        ubuntu-upstream     \           \`-------G           \
                             \           \        \           \
        debian-packaging      \ ,---------D--------\-----------J
                               C                    \
        ubuntu-packaging        `----E---------------I

        where:
           - A = 1.0
           - B = 1.1
           - H = 2.0

           - G = 1.1.2

           - C = 1.0-1
           - D = 1.1-1
           - J = 2.0-1

           - E = 1.0-1ubuntu1
           - I = 1.1.2-0ubuntu1

        Please note that the debian upstream and the ubuntu packaging
        branches will have a conflict with respect to the file 'c'.
        """
        # Set up the debian upstream branch.
        name = 'debu-n'
        vdata = [
            ('upstream-1.0', ('a',), None, None),
            ('upstream-1.1', ('b',), None, None),
            ('upstream-2.0', ('c',), None, None)]
        debu_n = self._setup_branch(name, vdata)

        # Set up the debian packaging branch.
        name = 'debp-n'
        debp_n = self.make_branch_and_tree(name)
        debp_n.pull(debu_n.branch, stop_revision=self.revid_debu_n_A)

        vdata = [
            ('1.0-1', ('debian/', 'debian/changelog'), None, None),
            ('1.1-1', ('o',), debu_n, self.revid_debu_n_B),
            ('2.0-1', ('p',), debu_n, self.revid_debu_n_C)]
        self._setup_branch(name, vdata, debp_n, 'd')

        # Set up the ubuntu upstream branch.
        name = 'ubuu-o'
        ubuu_o = debu_n.bzrdir.sprout(
            name, revision_id=self.revid_debu_n_B).open_workingtree()

        vdata = [('upstream-1.1.2', (), None, None)]
        self._setup_branch(name, vdata, ubuu_o)

        # Set up the ubuntu packaging branch.
        name = 'ubup-o'
        ubup_o = debu_n.bzrdir.sprout(
            name, revision_id=self.revid_debu_n_A).open_workingtree()

        vdata = [
            ('1.0-1ubuntu1', (), debp_n, self.revid_debp_n_A),
            ('1.1.2-0ubuntu1', ('c',), ubuu_o, self.revid_ubuu_o_A)]
        self._setup_branch(name, vdata, ubup_o, 'u')

        # Return the ubuntu and the debian packaging branches.
        return (ubup_o, debp_n)

    def _setup_branch(self, name, vdata, tree=None, log_format=None):
        vids = list(string.ascii_uppercase)
        days = range(len(string.ascii_uppercase))

        if tree is None:
            tree = self.make_branch_and_tree(name)

        def revid_name(vid):
            return 'revid_%s_%s' % (name.replace('-', '_'), vid)

        def add_paths(paths):
            qpaths = ['%s/%s' % (name, path) for path in paths]
            self.build_tree(qpaths)
            tree.add(paths)

        def changelog(vdata, vid):
            result = ''
            day = days.pop(0)
            if isinstance(vdata, tuple):
                uver, dver = vdata[:2]
                ucle = _Ubuntu_changelog % (uver, vid, day)
                dcle = _Debian_changelog % (dver, vid, day)
                result = ucle + dcle
            else:
                if log_format == 'u':
                    result = _Ubuntu_changelog % (vdata, vid, day)
                elif log_format == 'd':
                    result = _Debian_changelog % (vdata, vid, day)

            return result

        def commit(msg, version):
            vid = vids.pop(0)
            if log_format is not None:
                cle = changelog(version, vid)
                p = '%s/work/%s/debian/changelog' % (self.test_base_dir, name)
                _prepend_log(cle, p)
            revid = tree.commit('%s: %s' % (vid, msg))
            setattr(self, revid_name(vid), revid)
            tree.branch.tags.set_tag(version, revid)

        def tree_nick(tree):
            return str(tree)[1:-1].split('/')[-1]
            
        tree.lock_write()

        try:
            for version, paths, utree, urevid in vdata:
                msg = ''
                if utree is not None:
                    tree.merge_from_branch(utree.branch, to_revision=urevid)
                    utree.branch.tags.merge_to(tree.branch.tags)
                    if urevid is not None:
                        msg += 'Merged tree %s|%s. ' % (tree_nick(utree), urevid)
                    else:
                        msg += 'Merged tree %s. ' % utree
                if paths is not None:
                    add_paths(paths)
                    msg += 'Added paths: %s. ' % str(paths)

                commit(msg, version)
        finally:
            tree.unlock()

        return tree



#    source_distiller.py -- Getting the source to build from a branch
#    Copyright (C) 2008, 2009 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import glob
import os
import shutil
import tempfile

from bzrlib import errors as bzr_errors

from bzrlib.plugins.builddeb.util import (
    export,
    extract_orig_tarballs,
    get_parent_dir,
    recursive_copy,
    )


class SourceDistiller(object):
    """A source distiller extracts the source to build from a location.

    It does whatever is needed to give you a source you can build at
    a location of your choice.
    """

    supports_use_existing = False

    def __init__(self, tree, upstream_provider, top_level=False,
            use_existing=False, is_working_tree=False):
        """Create a SourceDistiller to distill from the specified tree.

        :param tree: The tree to use as the source.
        :param upstream_provider: an UpstreamProvider to provide the upstream
            tarball if needed.
        :param top_level: if the tree is in the top level directory instead of inside debian/.
        :param use_existing: whether the distiller should re-use an existing
            target if the distiller supports it.
        :param is_working_tree: if `tree` is a working tree.
        """
        self.tree = tree
        self.upstream_provider = upstream_provider
        self.top_level = top_level
        self.use_existing = use_existing
        if not self.supports_use_existing:
            assert not self.use_existing, "distiller doesn't support use_existing"
        self.is_working_tree = is_working_tree

    def distill(self, target):
        """Extract the source to a tree rooted at the given location.

        The passed location cannot already exist. If it does then
        FileExists will be raised.

        :param target: a string containing the location at which to 
            place the tree containing the buildable source.
        """
        if not self.supports_use_existing or not self.use_existing:
            if os.path.exists(target):
                raise bzr_errors.FileExists(target)
        elif self.supports_use_existing and self.use_existing:
            if not os.path.exists(target):
                raise bzr_errors.NotADirectory(target)
        self._distill(target)

    def _distill(self, target):
        """Subclasses should override this to implement distill."""
        raise NotImplementedError(self._distill)

    def _prepare_working_tree(self):
        for (dp, ie) in self.tree.iter_entries_by_dir():
            ie._read_tree_state(dp, self.tree)


class NativeSourceDistiller(SourceDistiller):
    """A SourceDistiller for unpacking a native package from a branch."""

    def _distill(self, target):
        if self.is_working_tree:
            self._prepare_working_tree()
        export(self.tree, target, None, None)


class FullSourceDistiller(SourceDistiller):
    """A SourceDistiller for full-source branches, a.k.a. normal mode"""

    def _distill(self, target):
        parent_dir = get_parent_dir(target)
        self.upstream_provider.provide(parent_dir)
        if self.is_working_tree:
            self._prepare_working_tree()
        export(self.tree, target)


class MergeModeDistiller(SourceDistiller):

    supports_use_existing = True

    def _distill(self, target):
        # Get the upstream tarball
        parent_dir = get_parent_dir(target)
        if parent_dir != '' and not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        if not self.use_existing:
            tarballs = self.upstream_provider.provide(parent_dir)
            # Extract it to the right place
            tempdir = tempfile.mkdtemp(prefix='builddeb-merge-')
            try:
                extract_orig_tarballs(tarballs, tempdir)
                files = glob.glob(tempdir+'/*')
                # If everything is in a single dir then move everything up one
                # level.
                if os.path.isdir(target):
                    shutil.rmtree(target)
                if len(files) == 1:
                    shutil.move(files[0], target)
                else:
                    shutil.move(tempdir, target)
            finally:
                if os.path.exists(tempdir):
                    shutil.rmtree(tempdir)
        # Now export the tree to provide the debian dir
        basetempdir = tempfile.mkdtemp(prefix='builddeb-merge-debian-')
        try:
            tempdir = os.path.join(basetempdir,"export")
            if self.top_level:
                os.makedirs(tempdir)
                export_dir = os.path.join(tempdir, 'debian')
            else:
                export_dir = tempdir
            if self.is_working_tree:
                self._prepare_working_tree()
            export(self.tree, export_dir)
            # Remove any upstream debian dir, or from previous export with
            # use_existing
            if os.path.exists(os.path.join(target, 'debian')):
                shutil.rmtree(os.path.join(target, 'debian'))
            recursive_copy(tempdir, target)
        finally:
            shutil.rmtree(basetempdir)



#    test_merge_upstream.py -- Testsuite for builddeb's upstream merging.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

try:
    from debian.changelog import Changelog, Version
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle.changelog import Changelog, Version

from bzrlib.tests import (
    TestCase,
    TestCaseWithTransport,
    )

from bzrlib.plugins.builddeb.merge_upstream import (
    changelog_add_new_version,
    upstream_merge_changelog_line,
    package_version,
    )


class TestPackageVersion(TestCase):

    def test_simple_debian(self):
        self.assertEquals(Version("1.2-1"),
                package_version("1.2", "debian"))

    def test_simple_ubuntu(self):
        self.assertEquals(Version("1.2-0ubuntu1"),
                package_version("1.2", "ubuntu"))

    def test_debian_with_dash(self):
        self.assertEquals(Version("1.2-0ubuntu1-1"),
                package_version("1.2-0ubuntu1", "debian"))

    def test_ubuntu_with_dash(self):
        self.assertEquals(Version("1.2-1-0ubuntu1"),
                package_version("1.2-1", "ubuntu"))

    def test_ubuntu_with_epoch(self):
        self.assertEquals(Version("3:1.2-1-0ubuntu1"),
                package_version("1.2-1", "ubuntu", "3"))


class UpstreamMergeChangelogLineTests(TestCase):

    def test_release(self):
        self.assertEquals("New upstream release.", upstream_merge_changelog_line("1.0"))

    def test_bzr_snapshot(self):
        self.assertEquals("New upstream snapshot.",
            upstream_merge_changelog_line("1.0+bzr3"))

    def test_git_snapshot(self):
        self.assertEquals("New upstream snapshot.",
            upstream_merge_changelog_line("1.0~git20101212"))

    def test_plus(self):
        self.assertEquals("New upstream release.",
            upstream_merge_changelog_line("1.0+dfsg1"))


class ChangelogAddNewVersionTests(TestCaseWithTransport):

    def test_add_new(self):
        tree = self.make_branch_and_tree(".")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.mkdir("debian")
        changelog_add_new_version(tree, "1.0", "sid", None, "somepkg")
        # changelog_add_new_version will version the changelog if it was created
        cl = Changelog(open('debian/changelog'))
        self.assertEquals(cl._blocks[0].package, "somepkg")
        self.assertEquals(cl._blocks[0].distributions, "UNRELEASED")
        self.assertEquals(cl._blocks[0].version, Version("1.0-1"))
        self.assertEquals([], list(tree.filter_unversioned_files(["debian/changelog"])))



#    test_dep3.py -- Testsuite for builddeb dep3.py
#    Copyright (C) 2011 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from cStringIO import StringIO

import rfc822

from bzrlib.revision import (
    NULL_REVISION,
    )
from bzrlib.tests import (
    TestCase,
    )

from bzrlib.plugins.builddeb.dep3 import (
    describe_origin,
    determine_applied_upstream,
    gather_bugs_and_authors,
    write_dep3_bug_line,
    write_dep3_patch,
    write_dep3_patch_header,
    )

from bzrlib.plugins.builddeb.tests import (
    TestCaseWithTransport,
    )


class Dep3HeaderTests(TestCase):

    def dep3_header(self, description=None, origin=None, forwarded=None,
            bugs=None, authors=None, revision_id=None, last_update=None,
            applied_upstream=None):
        f = StringIO()
        write_dep3_patch_header(f, description=description, origin=origin,
            forwarded=forwarded, bugs=bugs, authors=authors,
            revision_id=revision_id, last_update=last_update,
            applied_upstream=applied_upstream)
        f.seek(0)
        return rfc822.Message(f)

    def test_description(self):
        ret = self.dep3_header(description="This patch fixes the foobar")
        self.assertEquals("This patch fixes the foobar", ret["Description"])

    def test_last_updated(self):
        ret = self.dep3_header(last_update=1304840034)
        self.assertEquals("2011-05-08", ret["Last-Update"])

    def test_revision_id(self):
        ret = self.dep3_header(revision_id="myrevid")
        self.assertEquals("myrevid", ret["X-Bzr-Revision-Id"])

    def test_authors(self):
        authors = [
            "Jelmer Vernooij <jelmer@canonical.com>",
            "James Westby <james.westby@canonical.com>"]
        ret = self.dep3_header(authors=authors)
        self.assertEquals([
            ("Jelmer Vernooij", "jelmer@canonical.com"),
            ("James Westby", "james.westby@canonical.com")],
            ret.getaddrlist("Author"))

    def test_origin(self):
        ret = self.dep3_header(origin="Cherrypick from upstream")
        self.assertEquals("Cherrypick from upstream",
            ret["Origin"])

    def test_forwarded(self):
        ret = self.dep3_header(forwarded="not needed")
        self.assertEquals("not needed",
            ret["Forwarded"])

    def test_applied_upstream(self):
        ret = self.dep3_header(applied_upstream="commit 45")
        self.assertEquals("commit 45", ret["Applied-Upstream"])

    def test_bugs(self):
        bugs = [
            ("http://bugs.debian.org/424242", "fixed"),
            ("https://bugs.launchpad.net/bugs/20110508", "fixed"),
            ("http://bugzilla.samba.org/bug.cgi?id=52", "fixed")]
        ret = self.dep3_header(bugs=bugs)
        self.assertEquals([
            "https://bugs.launchpad.net/bugs/20110508",
            "http://bugzilla.samba.org/bug.cgi?id=52"],
            ret.getheaders("Bug"))
        self.assertEquals(["http://bugs.debian.org/424242"],
            ret.getheaders("Bug-Debian"))

    def test_write_bug_fix_only(self):
        # non-fixed bug lines are ignored
        f = StringIO()
        write_dep3_bug_line(f, "http://bar/", "pending")
        self.assertEquals("", f.getvalue())

    def test_write_normal_bug(self):
        f = StringIO()
        write_dep3_bug_line(f, "http://bugzilla.samba.org/bug.cgi?id=42",
            "fixed")
        self.assertEquals("Bug: http://bugzilla.samba.org/bug.cgi?id=42\n",
            f.getvalue())

    def test_write_debian_bug(self):
        f = StringIO()
        write_dep3_bug_line(f, "http://bugs.debian.org/234354", "fixed")
        self.assertEquals("Bug-Debian: http://bugs.debian.org/234354\n",
            f.getvalue())


class GatherBugsAndAuthors(TestCaseWithTransport):

    def test_none(self):
        branch = self.make_branch(".")
        self.assertEquals((set(), set(), None),
            gather_bugs_and_authors(branch.repository, []))

    def test_multiple_authors(self):
        tree = self.make_branch_and_tree(".")
        revid1 = tree.commit(authors=["Jelmer Vernooij <jelmer@canonical.com>"],
                timestamp=1304844311, message="msg")
        revid2 = tree.commit(authors=["Max Bowsher <maxb@f2s.com>"],
                timestamp=1304844278, message="msg")
        self.assertEquals((set(), set([
            "Jelmer Vernooij <jelmer@canonical.com>",
            "Max Bowsher <maxb@f2s.com>"]), 1304844311),
            gather_bugs_and_authors(tree.branch.repository, [revid1, revid2]))

    def test_bugs(self):
        tree = self.make_branch_and_tree(".")
        revid1 = tree.commit(authors=["Jelmer Vernooij <jelmer@canonical.com>"],
                timestamp=1304844311, message="msg", revprops={"bugs":
                    "http://bugs.samba.org/bug.cgi?id=2011 fixed\n"})
        self.assertEquals((
            set([("http://bugs.samba.org/bug.cgi?id=2011", "fixed")]),
            set(["Jelmer Vernooij <jelmer@canonical.com>"]), 1304844311),
            gather_bugs_and_authors(tree.branch.repository, [revid1]))


class DetermineAppliedUpstreamTests(TestCaseWithTransport):

    def test_not_applied(self):
        upstream = self.make_branch_and_tree("upstream")
        feature = self.make_branch_and_tree("feature")
        feature.commit(message="every bloody emperor")
        self.addCleanup(feature.lock_read().unlock)
        self.assertEquals("no",
            determine_applied_upstream(upstream.branch, feature.branch))

    def test_merged(self):
        upstream = self.make_branch_and_tree("upstream")
        upstream.commit(message="initial upstream commit")
        feature = upstream.bzrdir.sprout("feature").open_workingtree()
        feature.commit(message="nutter alert")
        upstream.merge_from_branch(feature.branch)
        upstream.commit(message="merge feature")
        self.addCleanup(upstream.lock_read().unlock)
        self.addCleanup(feature.lock_read().unlock)
        self.assertEquals("merged in revision 2",
            determine_applied_upstream(upstream.branch, feature.branch))


class DescribeOriginTests(TestCaseWithTransport):

    def test_no_public_branch(self):
        tree = self.make_branch_and_tree(".")
        revid1 = tree.commit(message="msg1")
        self.assertEquals("commit, revision id: %s" % revid1,
            describe_origin(tree.branch, revid1))

    def test_public_branch(self):
        tree = self.make_branch_and_tree(".")
        tree.branch.set_public_branch("http://example.com/public")
        revid1 = tree.commit(message="msg1")
        self.assertEquals("commit, http://example.com/public, revision: 1",
            describe_origin(tree.branch, revid1))


class FullDep3PatchTests(TestCaseWithTransport):

    def test_simple(self):
        f = StringIO()
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("foo", "data")])
        tree.add("foo")
        revid = tree.commit("msg", rev_id="arevid", timestamp=1304849661,
            timezone=0)
        write_dep3_patch(f, tree.branch, NULL_REVISION, revid,
            description="Nutter alert",
            forwarded="not needed",
            authors=set(["Jelmer <jelmer@samba.org>"]))
        self.assertEquals("Description: Nutter alert\n"
            "Forwarded: not needed\n"
            "Author: Jelmer <jelmer@samba.org>\n"
            "X-Bzr-Revision-Id: arevid\n"
            "\n"
            "=== added file 'foo'\n"
            "--- old/foo\t1970-01-01 00:00:00 +0000\n"
            "+++ new/foo\t2011-05-08 10:14:21 +0000\n"
            "@@ -0,0 +1,1 @@\n"
            "+data\n"
            "\\ No newline at end of file\n"
            "\n", f.getvalue())



#    directory.py -- Directory service that uses Debian Vcs-* fields
#    Copyright (C) 2008 Jelmer Vernooij <jelmer@samba.org>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from bzrlib import errors
from bzrlib.trace import note

import apt_pkg


class VcsDirectory(object):
    """Simple Bazaar directory service which uses dpkg Vcs-* fields."""

    def look_up(self, name, url):
        if "/" in name:
            (name, version) = name.split("/", 1)
        else:
            version = None

        apt_pkg.init()

        # Older versions of apt_pkg don't have SourceRecords,
        # newer versions give a deprecation warning when using
        # GetPkgSrcRecords.
        try:
            sources = apt_pkg.SourceRecords()
        except AttributeError:
            sources = apt_pkg.GetPkgSrcRecords()

        urls = {}
        lookup = getattr(sources, 'lookup', None) or sources.Lookup
        while lookup(name):
            record = getattr(sources, 'record', None) or sources.Record
            for l in record.splitlines():
                if not ": " in l:
                    continue
                (field, value) = l.strip("\n").split(": ", 1)

                if field == "Version":
                    pkg_version = value
                elif field.startswith("X-Vcs-") or field.startswith("Vcs-"):
                    vcs = field.split("-")[-1]
                    urls.setdefault(pkg_version,{})[vcs] = value

        if len(urls) == 0:
            raise errors.InvalidURL(path=url, extra='no URLs found')

        if version is None:
            # Try the latest version
            cmp = getattr(apt_pkg, 'version_compare',
                    getattr(apt_pkg, 'VersionCompare', None))
            version = sorted(urls,cmp=cmp)[0]

        if not version in urls:
            raise errors.InvalidURL(path=url,
                    extra='version %s not found' % version)
        
        note("Retrieving Vcs locating from %s Debian version %s", name, version)

        if "Bzr" in urls[version]:
            return urls[version]["Bzr"]

        if "Svn" in urls[version]:
            try:
                import bzrlib.plugins.svn
            except ImportError:
                note("This package uses subversion. If you would like to "
                        "access it with bzr then please install bzr-svn "
                        "and re-run the command.")
            else:
                return urls[version]["Svn"]

        if "Git" in urls[version]:
            try:
                import bzrlib.plugins.git
            except ImportError:
                note("This package uses git. If you would like to "
                        "access it with bzr then please install bzr-git "
                        "and re-run the command.")
            else:
                return urls[version]["Git"]

        if "Hg" in urls[version]:
            try:
                import bzrlib.plugins.hg
            except ImportError:
                note("This package uses hg. If you would like to "
                        "access it with bzr then please install bzr-hg"
                        "and re-run the command.")
            else:
                return urls[version]["Hg"]

        raise errors.InvalidURL(path=url,
            extra='unsupported VCSes %r found' % urls[version].keys())


def upstream_branch_alias(b):
    from bzrlib.directory_service import directories
    from bzrlib.plugins.builddeb.util import debuild_config
    b.lock_read()
    try:
        tree = b.basis_tree()
        config = debuild_config(tree, False)
        return directories.dereference(config.upstream_branch)
    finally:
        b.unlock()




#    test_get_tar.py -- Blackbox tests for get-orig-source.
#    Copyright 2011 Canonical Ltd
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os

try:
    from debian.changelog import (Changelog,
                                  Version,
                                  )
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle.changelog import (Changelog,
                                         Version,
                                        )


from bzrlib.plugins.builddeb.tests import BuilddebTestCase


class TestGetOrigSource(BuilddebTestCase):

    def make_changelog(self, version=None):
        if version is None:
            version = self.package_version
        c = Changelog()
        c.new_block()
        c.version = Version(version)
        c.package = self.package_name
        c.distributions = 'unstable'
        c.urgency = 'low'
        c.author = 'James Westby <jw+debian@jameswestby.net>'
        c.date = 'The,  3 Aug 2006 19:16:22 +0100'
        c.add_change('')
        c.add_change('  *  test build')
        c.add_change('')
        return c

    def make_unpacked_source(self):
        """
        Create an unpacked source tree in a branch. Return the working
        tree
        """
        tree = self.make_branch_and_tree('.')
        cl_file = 'debian/changelog'
        source_files = ['README', 'debian/'] + [cl_file]
        self.build_tree(source_files)
        c = self.make_changelog()
        self.write_changelog(c, cl_file)
        tree.add(source_files)
        return tree

    def make_source_with_upstream(self, path="."):
        """Create a source tree in a branch with an upstream tag."""
        tree = self.make_branch_and_tree(path)
        source_files = ['README']
        self.build_tree([os.path.join(path, f) for f in source_files])
        tree.add(source_files)
        tree.commit("one", rev_id='revid1')
        tree.branch.tags.set_tag("upstream-0.1", tree.branch.last_revision())

        os.mkdir(os.path.join(path, 'debian'))
        c = self.make_changelog()
        self.write_changelog(c, os.path.join(path, 'debian/changelog'))
        tree.add(['debian', 'debian/changelog'])
        tree.commit("two", rev_id='revid2')
        return tree

    def test_get_orig_source_registered(self):
        self.run_bzr("get-orig-source --help")

    def test_get_orig_source_error_no_changelog(self):
        self.run_bzr_error(
        ['Could not find changelog at .*debian/changelog or .*changelog.'],
        "get-orig-source")

    def test_get_orig_source_error_no_tar(self):
        self.make_unpacked_source()
        self.run_bzr_error(
            ['Unable to find the needed upstream tarball for package test, '\
            'version 0.1.'],
            "get-orig-source")

    def test_get_orig_source(self):
        tree = self.make_source_with_upstream()
        self.run_bzr(['get-orig-source'])
        self.assertPathExists('../test_0.1.orig.tar.gz')

    def test_get_orig_source_directory(self):
        tree = self.make_source_with_upstream("somedir")
        self.run_bzr(['get-orig-source', '-d', 'somedir'])
        self.assertPathExists('../test_0.1.orig.tar.gz')

    def test_get_orig_source_explicit_version(self):
        tree = self.make_source_with_upstream()
        c = self.make_changelog("0.3-1")
        self.write_changelog(c, 'debian/changelog')
        tree.commit("package 0.3")
        self.run_bzr(['get-orig-source', '0.1'])
        self.assertPathExists('../test_0.1.orig.tar.gz')

    def test_get_orig_source_explicit_version_not_found(self):
        tree = self.make_source_with_upstream()
        self.run_bzr_error([
            'bzr: ERROR: Unable to find the needed upstream tarball for package test, version 0.3.'],
            'get-orig-source 0.3')



#    merge_package.py -- The plugin for bzr
#    Copyright (C) 2009 Canonical Ltd.
#
#    :Author: Muharem Hrnjadovic <muharem@ubuntu.com>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from bzrlib import version_info as bzrlib_version

import os
import shutil
import tempfile

try:
    from debian.changelog import Version
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle.changelog import Version

from bzrlib.plugins.builddeb.errors import (
    MultipleUpstreamTarballsNotSupported,
    SharedUpstreamConflictsWithTargetPackaging,
    )
from bzrlib.plugins.builddeb.import_dsc import DistributionBranch
from bzrlib.plugins.builddeb.util import find_changelog


def _upstream_version_data(branch, revid):
    """Most recent upstream versions/revision IDs of the merge source/target.

    Please note: both packaging branches must have been read-locked
    beforehand.

    :param branch: The merge branch.
    :param revid: The revision in the branch to consider
    :param tree: Optional tree for the revision
    """
    db = DistributionBranch(branch, branch)
    tree = branch.repository.revision_tree(revid)
    changelog, _ignore = find_changelog(tree, False)
    uver = changelog.version.upstream_version
    upstream_revids = db.pristine_upstream_source.version_as_revisions(None, uver)
    if upstream_revids.keys() != [None]:
        raise MultipleUpstreamTarballsNotSupported()
    upstream_revid = upstream_revids[None]
    return (Version(uver), upstream_revid)


def fix_ancestry_as_needed(tree, source, source_revid=None):
    """Manipulate the merge target's ancestry to avoid upstream conflicts.

    Merging J->I given the following ancestry tree is likely to result in
    upstream merge conflicts:

    debian-upstream                 ,------------------H
                       A-----------B                    \
    ubuntu-upstream     \           \`-------G           \
                         \           \        \           \
    debian-packaging      \ ,---------D--------\-----------J
                           C           \        \
    ubuntu-packaging        `----E------F--------I

    Here there was a new upstream release (G) that Ubuntu packaged (I), and
    then another one that Debian packaged, skipping G, at H and J.

    Now, the way to solve this is to introduce the missing link.

    debian-upstream                 ,------------------H------.
                       A-----------B                    \      \
    ubuntu-upstream     \           \`-------G-----------\------K
                         \           \        \           \
    debian-packaging      \ ,---------D--------\-----------J
                           C           \        \
    ubuntu-packaging        `----E------F--------I

    at K, which isn't a real merge, as we just use the tree from H, but add
    G as a parent and then we merge that in to Ubuntu.

    debian-upstream                 ,------------------H------.
                       A-----------B                    \      \
    ubuntu-upstream     \           \`-------G-----------\------K
                         \           \        \           \      \
    debian-packaging      \ ,---------D--------\-----------J      \
                           C           \        \                  \
    ubuntu-packaging        `----E------F--------I------------------L

    At this point we can merge J->L to merge the Debian and Ubuntu changes.

    :param tree: The `WorkingTree` of the merge target branch.
    :param source: The merge source (packaging) branch.
    """
    upstreams_diverged = False
    t_upstream_reverted = False
    target = tree.branch

    source.lock_read()
    try:
        if source_revid is None:
            source_revid = source.last_revision()
        tree.lock_write()
        try:
            # "Unpack" the upstream versions and revision ids for the merge
            # source and target branch respectively.
            (us_ver, us_revid) = _upstream_version_data(source, source_revid)
            (ut_ver, ut_revid) = _upstream_version_data(target,
                target.last_revision())

            # Did the upstream branches of the merge source/target diverge?
            graph = source.repository.get_graph(target.repository)
            upstreams_diverged = (len(graph.heads([us_revid, ut_revid])) > 1)

            # No, we're done!
            if not upstreams_diverged:
                return (upstreams_diverged, t_upstream_reverted)

            # Instantiate a `DistributionBranch` object for the merge target
            # (packaging) branch.
            db = DistributionBranch(tree.branch, tree.branch)
            tempdir = tempfile.mkdtemp(dir=os.path.join(tree.basedir, '..'))

            try:
                # Extract the merge target's upstream tree into a temporary
                # directory.
                db.extract_upstream_tree({None: ut_revid}, tempdir)
                tmp_target_utree = db.pristine_upstream_tree

                # Merge upstream branch tips to obtain a shared upstream parent.
                # This will add revision K (see graph above) to a temporary merge
                # target upstream tree.
                tmp_target_utree.lock_write()
                try:
                    if us_ver > ut_ver:
                        # The source upstream tree is more recent and the
                        # temporary target tree needs to be reshaped to match it.
                        tmp_target_utree.revert(
                            None, source.repository.revision_tree(us_revid))
                        t_upstream_reverted = True

                    tmp_target_utree.set_parent_ids((ut_revid, us_revid))
                    new_revid = tmp_target_utree.commit(
                        'Prepared upstream tree for merging into target branch.')

                    # Repository updates during a held lock are not visible,
                    # hence the call to refresh the data in the /target/ repo.
                    tree.branch.repository.refresh_data()

                    tree.branch.fetch(source, last_revision=us_revid)
                    tree.branch.fetch(tmp_target_utree.branch,
                            last_revision=new_revid)

                    # Merge shared upstream parent into the target merge branch. This
                    # creates revison L in the digram above.
                    conflicts = tree.merge_from_branch(tmp_target_utree.branch)
                    if conflicts > 0:
                        if bzrlib_version >= (2, 5):
                            cmd = "bzr merge"
                        else:
                            cmd = "bzr merge-package"
                        raise SharedUpstreamConflictsWithTargetPackaging(cmd)
                    else:
                        tree.commit('Merging shared upstream rev into target branch.')

                finally:
                    tmp_target_utree.unlock()
            finally:
                shutil.rmtree(tempdir)
        finally:
            tree.unlock()
    finally:
        source.unlock()

    return (upstreams_diverged, t_upstream_reverted)



#    test_comit_message.py -- Test hook for pre-filling commit message.
#    Copyright (C) 2009 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from bzrlib.plugins.builddeb import debian_changelog_commit_message, debian_changelog_commit
from bzrlib.plugins.builddeb.tests import TestCaseWithTransport
try:
    from bzrlib.tests.features import Feature
except ImportError: # bzr < 2.4
    from bzrlib.tests import Feature


class _LaunchpadConnectionFeature(Feature):

    def _probe(self):
        try:
            from httplib2 import Http, ServerNotFoundError
        except ImportError:
            return False
        try:
            Http().request("https://code.launchpad.net/")
        except ServerNotFoundError:
            return False
        return True


LaunchpadConnectionFeature = _LaunchpadConnectionFeature()


class CommitMessageTests(TestCaseWithTransport):

    class _Commit(object):
        class _Builder(object):
            _revprops = {}
        def __init__(self, work_tree, exclude=[], specific_files=[]):
            self.work_tree = work_tree
            self.exclude = exclude
            self.specific_files = specific_files
            self.builder = self._Builder()

    def set_changelog_content(self, content):
        f = open("debian/changelog", 'wb')
        try:
            f.write(content)
        finally:
            f.close()

    def _set_commit_message_from_changelog(self, value):
        f = open("debian/bzr-builddeb.conf", 'wb')
        try:
            f.write("[BUILDDEB]\ncommit-message-from-changelog = %s" % value)
        finally:
            f.close()

    def test_leaves_existing_message(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/'])
        self._set_commit_message_from_changelog(True)
        wt.add(['a', 'debian', 'debian/bzr-builddeb.conf'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit_message(commit, "foo"), "foo")

    def test_ignores_commit_without_debian_changelog(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/'])
        self._set_commit_message_from_changelog(True)
        wt.add(['a', 'debian', 'debian/bzr-builddeb.conf'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit_message(commit, None), None)

    def test_ignores_commit_excluding_debian_changelog(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(True)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * new line")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt, exclude=["debian/changelog"])
        self.assertEqual(debian_changelog_commit_message(commit, None), None)

    def test_ignores_commit_specific_files(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(True)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * new line\n")
        wt.add(['a'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt, specific_files=["a"])
        self.assertEqual(debian_changelog_commit_message(commit, None), None)

    def test_provides_stripped_message(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(True)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * new line\n")
        wt.add(['a'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit_message(commit, None),
                "new line\n")

    def test_provides_unstripped_message(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(True)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * two\n  * changes\n")
        wt.add(['a'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit_message(commit, None),
                "* two\n* changes\n")

    def test_no_set_message_config_option(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(False)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * a change\n")
        wt.add(['a'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit(commit, None),
                None)

    def test_set_message_with_bugs(self):
        self.requireFeature(LaunchpadConnectionFeature)
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(True)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * fix LP: #1234\n  * close LP: #4321\n")
        wt.add(['a'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit(commit, None),
                "* fix LP: #1234\n* close LP: #4321\n")
        self.assertEqual(commit.builder._revprops, 
                        {'bugs': 'https://launchpad.net/bugs/1234 fixed\n'
                                 'https://launchpad.net/bugs/4321 fixed'})

    def test_set_message_returns_unicode(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a', 'debian/', 'debian/changelog'])
        self._set_commit_message_from_changelog(True)
        wt.add(['debian/', 'debian/changelog', 'debian/bzr-builddeb.conf'])
        wt.commit("one")
        self.set_changelog_content("  * \xe2\x80\xa6real fix this time\n")
        wt.add(['a'])
        wt.lock_read()
        self.addCleanup(wt.unlock)
        commit = self._Commit(wt)
        self.assertEqual(debian_changelog_commit(commit, None),
                u"\u2026real fix this time\n")



#    merge_upstream.py -- Merge new upstream versions of packages.
#    Copyright (C) 2007 Reinhard Tartler <siretart@tauware.de>
#                  2007 James Westby <jw+debian@jameswestby.net>
#                  2008 Jelmer Vernooij <jelmer@samba.org>
#
#    Code is also taken from bzrtools, which is
#             (C) 2005, 2006, 2007 Aaron Bentley <aaron.bentley@utoronto.ca>
#             (C) 2005, 2006 Canonical Limited.
#             (C) 2006 Michael Ellerman.
#    and distributed under the GPL, version 2 or later.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import subprocess

try:
    from debian.changelog import Version
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle.changelog import Version

from bzrlib.plugins.builddeb.errors import (
    DchError,
    )


TAG_PREFIX = "upstream-"


def package_version(upstream_version, distribution_name, epoch=None):
    """Determine the package version for a new upstream.

    :param upstream_version: Upstream version string
    :param distribution_name: Distribution the package is for
    :param epoch: Optional epoch
    """
    assert isinstance(upstream_version, str), \
        "upstream_version should be a str, not %s" % str(
                type(upstream_version))
    if distribution_name == "ubuntu":
        ret = Version("%s-0ubuntu1" % upstream_version)
    else:
        ret = Version("%s-1" % upstream_version)
    ret.epoch = epoch
    return ret


def upstream_merge_changelog_line(upstream_version):
    """Describe that a new upstream revision was merged.

    This will either describe that a new upstream release or a new upstream snapshot
    was merged.

    :param upstream_version: Upstream version string
    :return: Line string for use in changelog
    """
    vcs_suffixes = ["~bzr", "+bzr", "~svn", "+svn", "~git", "+git"]
    for vcs_suffix in vcs_suffixes:
        if vcs_suffix in str(upstream_version):
            entry_description = "New upstream snapshot."
            break
    else:
        entry_description = "New upstream release."
    return entry_description


def changelog_add_new_version(tree, upstream_version, distribution_name,
        changelog, package):
    """Add an entry to the changelog for a new version.

    :param tree: WorkingTree in which the package lives
    :param upstream_version: Upstream version to add
    :param distribution_name: Distribution name (debian, ubuntu, ...)
    :param changelog: Changelog object
    :param package: Package name
    """
    assert isinstance(upstream_version, str), \
         "upstream_version should be a str, not %s" % str(
                 type(upstream_version))
    entry_description = upstream_merge_changelog_line(upstream_version)
    if changelog is None:
        epoch = None
    else:
        epoch = changelog.epoch
    argv = ["dch", "-v",
            str(package_version(upstream_version, distribution_name, epoch)),
            "-D", "UNRELEASED", "--release-heuristic", "changelog",
            "--package", package, entry_description]
    create = (not tree.has_filename("debian/changelog"))
    if create:
        argv.append("--create")
    proc = subprocess.Popen(argv, cwd=tree.basedir, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    (stdout, stderr) = proc.communicate()
    if proc.returncode != 0:
        raise DchError("Adding changelog entry failed: %s" % stderr)
    if create:
        tree.add(["debian/changelog"])



#    test_config.py -- Tests for builddeb's config.py
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from bzrlib.branch import Branch

from bzrlib.plugins.builddeb.config import (
    BUILD_TYPE_MERGE,
    BUILD_TYPE_NORMAL,
    DebBuildConfig,
    )
from bzrlib.plugins.builddeb.tests import TestCaseWithTransport


class DebBuildConfigTests(TestCaseWithTransport):

  def setUp(self):
    super(DebBuildConfigTests, self).setUp()
    self.tree = self.make_branch_and_tree('.')
    self.branch = self.tree.branch
    f = open('default.conf', 'wb')
    try:
      f.write('['+DebBuildConfig.section+']\n')
      f.write('builder = invalid builder\n') # shouldn't be read as it needs
                                             # to be trusted
      f.write('build-dir = default build dir\n')
      f.write('orig-dir = default orig dir\n')
      f.write('result-dir = default result dir\n')
    finally:
      f.close()
    f = open('user.conf', 'wb')
    try:
      f.write('['+DebBuildConfig.section+']\n')
      f.write('builder = valid builder\n')
      f.write('quick-builder = valid quick builder\n')
      f.write('orig-dir = user orig dir\n')
      f.write('result-dir = user result dir\n')
    finally:
      f.close()
    f = open('.bzr/branch/branch.conf', 'wb')
    try:
      f.write('['+DebBuildConfig.section+']\n')
      f.write('quick-builder = invalid quick builder\n')
      f.write('result-dir = branch result dir\n')
    finally:
      f.close()
    self.tree.add(['default.conf', 'user.conf'])
    self.config = DebBuildConfig([('user.conf', True),
                                  ('default.conf', False)], branch=self.branch)

  def test_secure_not_from_untrusted(self):
    self.assertEqual(self.config.builder, 'valid builder')

  def test_secure_not_from_branch(self):
    self.assertEqual(self.config.quick_builder, 'valid quick builder')

  def test_branch_over_all(self):
    self.assertEqual(self.config.result_dir, 'branch result dir')

  def test_hierarchy(self):
    self.assertEqual(self.config.orig_dir, 'user orig dir')
    self.assertEqual(self.config.build_dir, 'default build dir')

  def test_no_entry(self):
    self.assertEqual(self.config.merge, False)
    self.assertEqual(self.config.build_type, None)

  def test_parse_error(self):
    f = open('invalid.conf', 'wb')
    try:
      f.write('['+DebBuildConfig.section+'\n')
    finally:
      f.close()
    DebBuildConfig([('invalid.conf', True, 'invalid.conf')])

try:
  from bzrlib.plugins.svn.config import SubversionBuildPackageConfig
except ImportError:
  pass
else:
  from bzrlib.plugins.svn.tests import SubversionTestCase

  class DebuildSvnBpTests(SubversionTestCase):

    if not getattr(SubversionTestCase, "make_svn_branch", None):
      def make_svn_branch(self, relpath):
        repos_url = self.make_repository(relpath)
        return Branch.open(repos_url)

    def test_from_properties(self):
      branch = self.make_svn_branch("d")

      cfg = DebBuildConfig([], tree=branch.basis_tree())
      self.assertEquals(False, cfg.merge)

      dc = self.get_commit_editor(branch.base)
      d = dc.add_dir("debian")
      d.change_prop("mergeWithUpstream", "1")
      d.change_prop("svn-bp:origDir", "someorigdir")
      dc.close()

      cfg = DebBuildConfig([], tree=branch.basis_tree())
      self.assertEquals(True, cfg.merge)
      self.assertEquals(BUILD_TYPE_MERGE, cfg.build_type)
      self.assertEquals("someorigdir", cfg.orig_dir)

    def test_from_svn_layout_file(self):
      branch = self.make_svn_branch("d")

      cfg = DebBuildConfig([], tree=branch.basis_tree())
      self.assertEquals(False, cfg.merge)

      dc = self.get_commit_editor(branch.base)
      d = dc.add_dir("debian")
      f = d.add_file("debian/svn-layout")
      f.modify("origDir = someorigdir\n")
      dc.close()

      cfg = DebBuildConfig([], tree=branch.basis_tree())
      self.assertEquals("someorigdir", cfg.orig_dir)

# vim: ts=2 sts=2 sw=2



#    version.py -- Defines the version of builddeb.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

# same format as sys.version_info: "A tuple containing the five components of
# the version number: major, minor, micro, releaselevel, and serial. All
# values except releaselevel are integers; the release level is 'alpha',
# 'beta', 'candidate', or 'final'. The version_info value corresponding to the
# Python version 2.0 is (2, 0, 0, 'final', 0)."  Additionally we use a
# releaselevel of 'dev' for unreleased under-development code.
#
# Please set this to 'final' before upload.
version_info = (2, 8, 8, 'dev', 0)




#    __init__.py -- The plugin for bzr
#    Copyright (C) 2005 Jamie Wilkinson <jaq@debian.org> 
#                  2006, 2007 James Westby <jw+debian@jameswestby.net>
#                  2007 Reinhard Tartler <siretart@tauware.de>
#                  2008-2011 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os
import shutil
import tempfile

from bzrlib import (
    urlutils,
    version_info as bzrlib_version,
    )
from bzrlib.branch import Branch
from bzrlib.bzrdir import BzrDir
from bzrlib.commands import Command
from bzrlib.errors import (
    BzrCommandError,
    FileExists,
    NotBranchError,
    NoWorkingTree,
    )
from bzrlib.option import Option
from bzrlib.trace import note, warning
from bzrlib.workingtree import WorkingTree

from bzrlib.plugins.builddeb import (
    default_build_dir,
    default_orig_dir,
    default_result_dir,
    gettext,
    )
from bzrlib.plugins.builddeb.config import (
    BUILD_TYPE_MERGE,
    BUILD_TYPE_NATIVE,
    BUILD_TYPE_SPLIT,
    )
from bzrlib.plugins.builddeb.util import (
    debuild_config,
    )

dont_purge_opt = Option('dont-purge',
    help="Don't purge the build directory after building.")
result_opt = Option('result-dir',
    help="Directory in which to place the resulting package files.", type=str,
    argname="DIRECTORY")
builder_opt = Option('builder',
    help="Command to build the package.", type=str,
    argname="BUILDER")
merge_opt = Option('merge',
    help='Merge the debian part of the source in to the upstream tarball.')
split_opt = Option('split',
    help="Automatically create an .orig.tar.gz from a full source branch.")
build_dir_opt = Option('build-dir',
    help="The dir to use for building.", type=str)
orig_dir_opt = Option('orig-dir',
    help="Directory containing the .orig.tar.gz files. For use when only "
       +"debian/ is versioned.", type=str,
       argname="DIRECTORY")
native_opt = Option('native',
    help="Build a native package.")
export_upstream_opt = Option('export-upstream',
    help="Create the .orig.tar.gz from specified bzr branch before building.",
    type=unicode, argname="BRANCH")
export_upstream_revision_opt = Option('export-upstream-revision',
    help="Select the upstream revision that will be exported.",
    type=str, argname="REVISION")


def _get_changelog_info(tree, last_version=None, package=None, distribution=None):
    from bzrlib.plugins.builddeb.util import (
        find_changelog,
        find_last_distribution,
        lookup_distribution,
        )
    from bzrlib.plugins.builddeb.errors import (
        MissingChangelogError,
        )
    DEFAULT_FALLBACK_DISTRIBUTION = "debian"
    current_version = last_version
    try:
        (changelog, top_level) = find_changelog(tree, False, max_blocks=2)
    except MissingChangelogError:
        top_level = False
        changelog = None
        if distribution is None:
            distribution = DEFAULT_FALLBACK_DISTRIBUTION
            note(gettext("No distribution specified, and no changelog, "
                         "assuming '%s'"), distribution)
    else:
        if last_version is None:
            current_version = changelog.version.upstream_version
        if package is None:
            package = changelog.package
        if distribution is None:
            distribution = find_last_distribution(changelog)
            if distribution is not None:
                note(gettext("Using distribution %s") % distribution)
            else:
                distribution = DEFAULT_FALLBACK_DISTRIBUTION
                note(gettext("No distribution specified, and no previous "
                             "distribution in changelog. Assuming '%s'"),
                             distribution)
    distribution = distribution.lower()
    distribution_name = lookup_distribution(distribution)
    if distribution_name is None:
        raise BzrCommandError(gettext("Unknown target distribution: %s") \
                    % distribution)
    return (current_version, package, distribution, distribution_name,
            changelog, top_level)



class cmd_builddeb(Command):
    """Builds a Debian package from a branch.

    If BRANCH is specified it is assumed that the branch you wish to build is
    located there. If it is not specified then the current directory is used.

    By default, if a working tree is found, it is used to build. Otherwise the
    last committed revision found in the branch is used. To force building the
    last committed revision use --revision -1. You can also specify any other
    revision with the --revision option.

    If you only wish to export the package, and not build it (especially useful
    for merge mode), use --export-only.

    To leave the build directory when the build is completed use --dont-purge.

    Specify the command to use when building using the --builder option, by
    default "debuild" is used. It can be overriden by setting the "builder"
    variable in you configuration. You can specify extra options to build with
    by adding them to the end of the command, after using "--" to indicate the
    end of the options to builddeb itself. The builder that you specify must
    accept the options you provide at the end of its command line.

    You can also specify directories to use for different things. --build-dir
    is the directory to build the packages beneath, which defaults to
    '../build-area'. '--orig-dir' specifies the directory that contains the
    .orig.tar.gz files , which defaults to '..'. '--result-dir' specifies where
    the resulting package files should be placed, which defaults to '..'.
    --result-dir will have problems if you use a build command that places
    the results in a different directory.

    The --reuse option will be useful if you are in merge mode, and the upstream
    tarball is very large. It attempts to reuse a build directory from an earlier
    build. It will fail if one doesn't exist, but you can create one by using 
    --export-only. 

    --quick allows you to define a quick-builder in your configuration files, 
    which will be used when this option is passed. It defaults to 'fakeroot 
    debian/rules binary'. It is overriden if --builder is passed. Using this
    and --reuse allows for fast rebuilds.
    """
    working_tree_opt = Option('working-tree', help="This option has no effect.",
                              short_name='w')
    export_only_opt = Option('export-only', help="Export only, don't build.",
                             short_name='e')
    use_existing_opt = Option('use-existing',
                              help="Use an existing build directory.")
    quick_opt = Option('quick', help="Quickly build the package, uses "
                       +"quick-builder, which defaults to \"fakeroot "
                       +"debian/rules binary\".")
    reuse_opt = Option('reuse', help="Try to avoid exporting too much on each "
                       +"build. Only works in merge mode; it saves unpacking "
                       +"the upstream tarball each time. Implies --dont-purge "
                       +"and --use-existing.")
    source_opt = Option('source', help="Build a source package.",
                        short_name='S')
    strict_opt = Option('strict',
               help='Refuse to build if there are unknown files in'
               ' the working tree, --no-strict disables the check.')
    result_compat_opt = Option('result', help="Present only for compatibility "
            "with bzr-builddeb <= 2.0. Use --result-dir instead.")
    package_merge_opt = Option('package-merge', help="Build using the "
            "appropriate -v and -sa options for merging in the changes from "
            "another source.")
    takes_args = ['branch_or_build_options*']
    aliases = ['bd']
    takes_options = [working_tree_opt, export_only_opt,
        dont_purge_opt, use_existing_opt, result_opt, builder_opt, merge_opt,
        build_dir_opt, orig_dir_opt, split_opt,
        export_upstream_opt, export_upstream_revision_opt,
        quick_opt, reuse_opt, native_opt,
        source_opt, 'revision', strict_opt,
        result_compat_opt, package_merge_opt]

    def _get_tree_and_branch(self, location):
        import urlparse
        if location is None:
            location = "."
        is_local = urlparse.urlsplit(location)[0] in ('', 'file')
        bzrdir, relpath = BzrDir.open_containing(location)
        tree, branch = bzrdir._get_tree_branch()
        return tree, branch, is_local, bzrdir.user_url

    def _get_build_tree(self, revision, tree, branch):
        if revision is None and tree is not None:
            note(gettext("Building using working tree"))
            working_tree = True
        else:
            if revision is None:
                revid = branch.last_revision()
            elif len(revision) == 1:
                revid = revision[0].in_history(branch).rev_id
            else:
                raise BzrCommandError(gettext(
                    'bzr builddeb --revision takes exactly one '
                    'revision specifier.'))
            note(gettext("Building branch from revision %s"), revid)
            tree = branch.repository.revision_tree(revid)
            working_tree = False
        return tree, working_tree

    def _build_type(self, merge, native, split):
        if merge:
            return BUILD_TYPE_MERGE
        if native:
            return BUILD_TYPE_NATIVE
        if split:
            return BUILD_TYPE_SPLIT
        return None

    def _get_build_command(self, config, builder, quick, build_options):
        if builder is None:
            if quick:
                builder = config.quick_builder
                if builder is None:
                    builder = "fakeroot debian/rules binary"
            else:
                builder = config.builder
                if builder is None:
                    builder = "debuild"
        if build_options:
            builder += " " + " ".join(build_options)
        return builder

    def _get_dirs(self, config, location, is_local, result_dir, build_dir, orig_dir):
        def _get_dir(supplied, if_local, if_not):
            if supplied is None:
                if is_local:
                    supplied = if_local
                else:
                    supplied = if_not
            if supplied is not None:
                if is_local:
                    supplied = os.path.join(
                            urlutils.local_path_from_url(location),
                            supplied)
                    supplied = os.path.realpath(supplied)
            return supplied

        result_dir = _get_dir(result_dir, config.result_dir, config.user_result_dir)
        build_dir = _get_dir(build_dir, config.build_dir or default_build_dir,
                config.user_build_dir or 'build-area')
        orig_dir = _get_dir(orig_dir, config.orig_dir or default_orig_dir,
                config.user_orig_dir or 'build-area')
        return result_dir, build_dir, orig_dir

    def _branch_and_build_options(self, branch_or_build_options_list,
            source=False):
        branch = None
        build_options = []
        source_opt = False
        if branch_or_build_options_list is not None:
            for opt in branch_or_build_options_list:
                if opt.startswith("-") or branch is not None:
                    build_options.append(opt)
                    if opt == "-S" or opt == "--source":
                        source_opt = True
                else:
                    branch = opt
        if source and not source_opt:
            build_options.append("-S")
        if source_opt:
            source = True
        return branch, build_options, source

    def _get_upstream_branch(self, export_upstream, export_upstream_revision,
            config, version):
        from bzrlib.plugins.builddeb.upstream.branch import (
            LazyUpstreamBranchSource,
            )
        upstream_source = LazyUpstreamBranchSource(export_upstream,
            config=config)
        if export_upstream_revision:
            upstream_source.upstream_revision_map[version.encode("utf-8")] = export_upstream_revision
        return upstream_source

    def run(self, branch_or_build_options_list=None, verbose=False,
            working_tree=False,
            export_only=False, dont_purge=False, use_existing=False,
            result_dir=None, builder=None, merge=None, build_dir=None,
            export_upstream=None, export_upstream_revision=None,
            orig_dir=None, split=None,
            quick=False, reuse=False, native=None,
            source=False, revision=None, result=None, package_merge=None,
            strict=False):
        import commands
        from bzrlib.plugins.builddeb.source_distiller import (
                FullSourceDistiller,
                MergeModeDistiller,
                NativeSourceDistiller,
                )
        from bzrlib.plugins.builddeb.builder import DebBuild
        from bzrlib.plugins.builddeb.errors import (
            NoPreviousUpload,
            StrictBuildFailed
            )
        from bzrlib.plugins.builddeb.hooks import run_hook
        from bzrlib.plugins.builddeb.upstream.branch import (
            LazyUpstreamBranchSource,
            )
        from bzrlib.plugins.builddeb.upstream import (
            AptSource,
            DebianRulesSource,
            SelfSplitSource,
            UScanSource,
            UpstreamProvider,
            )
        from bzrlib.plugins.builddeb.upstream.pristinetar import (
            PristineTarSource,
            )
        from bzrlib.plugins.builddeb.util import (
            dget_changes,
            find_changelog,
            find_previous_upload,
            guess_build_type,
            tree_contains_upstream_source,
            )

        if result is not None:
            warning(gettext("--result is deprecated, use --result-dir instead"))
        location, build_options, source = self._branch_and_build_options(
                branch_or_build_options_list, source)
        tree, branch, is_local, location = self._get_tree_and_branch(location)
        tree, working_tree = self._get_build_tree(revision, tree, branch)
        if strict:
            for unknown in tree.unknowns():
                raise StrictBuildFailed()

        if len(tree.conflicts()) > 0:
            raise BzrCommandError(
                "There are conflicts in the working tree. "
                "You must resolve these before building.")

        tree.lock_read()
        try:
            config = debuild_config(tree, working_tree)
            if reuse:
                note(gettext("Reusing existing build dir"))
                dont_purge = True
                use_existing = True
            build_type = self._build_type(merge, native, split)
            if build_type is None:
                build_type = config.build_type
            contains_upstream_source = tree_contains_upstream_source(tree)
            (changelog, top_level) = find_changelog(tree, not contains_upstream_source)
            if build_type is None:
                build_type = guess_build_type(tree, changelog.version,
                    contains_upstream_source)

            note(gettext("Building package in %s mode") % build_type)

            if package_merge:
                try:
                    prev_version = find_previous_upload(tree,
                        not contains_upstream_source)
                except NoPreviousUpload:
                    prev_version = None
                if prev_version is None:
                    build_options.extend(["-sa", "-v0"])
                else:
                    build_options.append("-v%s" % str(prev_version))
                    if (prev_version.upstream_version != 
                            changelog.version.upstream_version or
                        prev_version.epoch != changelog.version.epoch):
                        build_options.append("-sa")
            build_cmd = self._get_build_command(config, builder, quick,
                    build_options)
            result_dir, build_dir, orig_dir = self._get_dirs(config,
                location or ".", is_local, result_dir or result, build_dir, orig_dir)

            upstream_sources = [
                PristineTarSource(tree, branch),
                AptSource(),
                ]
            if build_type == BUILD_TYPE_MERGE:
                if export_upstream is None and config.export_upstream:
                    export_upstream = config.export_upstream
                    warning(gettext("The 'export-upstream' configuration "
                        "option is deprecated. Use 'upstream-branch' instead."))
                if export_upstream is None and config.upstream_branch:
                    export_upstream = config.upstream_branch
                if export_upstream:
                    upstream_branch_source = self._get_upstream_branch(
                        export_upstream, export_upstream_revision, config,
                        changelog.version.upstream_version)
                    upstream_sources.append(upstream_branch_source)
            elif not native and config.upstream_branch is not None:
                upstream_sources.append(LazyUpstreamBranchSource(config.upstream_branch))
            upstream_sources.extend([
                DebianRulesSource(tree, top_level),
                UScanSource(tree, top_level),
                ])
            if build_type == BUILD_TYPE_SPLIT:
                upstream_sources.append(SelfSplitSource(tree))

            upstream_provider = UpstreamProvider(changelog.package,
                changelog.version.upstream_version, orig_dir, upstream_sources)

            if build_type == BUILD_TYPE_MERGE:
                distiller_cls = MergeModeDistiller
            elif build_type == BUILD_TYPE_NATIVE:
                distiller_cls = NativeSourceDistiller
            else:
                distiller_cls = FullSourceDistiller

            distiller = distiller_cls(tree, upstream_provider,
                    top_level=top_level, use_existing=use_existing,
                    is_working_tree=working_tree)

            build_source_dir = os.path.join(build_dir,
                    "%s-%s" % (changelog.package,
                               changelog.version.upstream_version))

            builder = DebBuild(distiller, build_source_dir, build_cmd,
                    use_existing=use_existing)
            builder.prepare()
            run_hook(tree, 'pre-export', config)
            builder.export()
            if not export_only:
                run_hook(tree, 'pre-build', config, wd=build_source_dir)
                builder.build()
                run_hook(tree, 'post-build', config, wd=build_source_dir)
                if not dont_purge:
                    builder.clean()
                if source:
                    arch = "source"
                else:
                    status, arch = commands.getstatusoutput(
                        'dpkg-architecture -qDEB_BUILD_ARCH')
                    if status > 0:
                        raise BzrCommandError("Could not find the build architecture")
                non_epoch_version = changelog.version.upstream_version
                if changelog.version.debian_version is not None:
                    non_epoch_version += "-%s" % changelog.version.debian_version
                changes = "%s_%s_%s.changes" % (changelog.package,
                        non_epoch_version, arch)
                changes_path = os.path.join(build_dir, changes)
                if not os.path.exists(changes_path):
                    if result_dir is not None:
                        raise BzrCommandError("Could not find the .changes "
                                "file from the build: %s" % changes_path)
                else:
                    if is_local:
                        target_dir = result_dir or default_result_dir
                        target_dir = os.path.join(
                                urlutils.local_path_from_url(location),
                                target_dir)
                    else:
                        target_dir = "."
                    if not os.path.exists(target_dir):
                        os.makedirs(target_dir)
                    dget_changes(changes_path, target_dir)
        finally:
            tree.unlock()


class cmd_get_orig_source(Command):
    """Gets the upstream tar file for the packaging branch."""

    directory_opt = Option('directory',
        help='Directory from which to retrieve the packaging data',
        short_name='d', type=unicode)

    takes_options = [directory_opt]
    takes_args = ["version?"]

    def run(self, directory='.', version=None):
        from bzrlib.plugins.builddeb.upstream import (
            AptSource,
            DebianRulesSource,
            UScanSource,
            UpstreamProvider,
            )
        from bzrlib.plugins.builddeb.upstream.pristinetar import (
            PristineTarSource,
            )
        from bzrlib.plugins.builddeb.util import (
            find_changelog,
            )
        tree = WorkingTree.open_containing(directory)[0]
        config = debuild_config(tree, tree)

        (changelog, larstiq) = find_changelog(tree, True)
        orig_dir = config.orig_dir
        if orig_dir is None:
            orig_dir = default_orig_dir

        if version is None:
            version = changelog.version.upstream_version

        upstream_provider = UpstreamProvider(changelog.package,
                str(version), orig_dir,
                [PristineTarSource(tree, tree.branch),
                 AptSource(),
                 DebianRulesSource(tree, larstiq),
                 UScanSource(tree, larstiq) ])

        result = upstream_provider.provide(orig_dir)
        for tar, component in result:
            note(gettext("Tar now in %s") % tar)


class cmd_merge_upstream(Command):
    """Merges a new upstream version into the current branch.

    Takes a new upstream version and merges it in to your branch, so that your
    packaging changes are applied to the new version.

    You must supply the source to import from, and in some cases
    the version number of the new release. The source can be a .tar.gz, .tar,
    .tar.bz2, .tar.xz, .tgz or .zip archive, a directory or a branch. The
    source may also be a remote file described by a URL.

    In most situations the version can be guessed from the upstream source.
    If the upstream version can not be guessed or if it is guessed
    incorrectly then the version number can be specified with --version.

    The distribution this version is targetted at can be specified with
    --distribution. This will be used to guess the version number suffix
    that you want, but you can always correct it in the resulting
    debian/changelog.

    If there is no debian changelog in the branch to retrieve the package
    name from then you must pass the --package option. If this version
    will change the name of the source package then you can use this option
    to set the new name.

    examples::

        bzr merge-upstream --version 0.2 \
            http://example.org/releases/scruff-0.2.tar.gz

    If you are merging a branch as well as the tarball then you can
    specify the branch after the tarball, along with -r to specify the
    revision of that branch to take::

        bzr merge-upstream --version 0.2 \
            http://example.org/releases/scruff-0.2.tar.gz \
            http://scruff.org/bzr/scruff.dev -r tag:0.2

    If there is no upstream release tarball, and you want bzr-builddeb to
    create the tarball for you::

        bzr merge-upstream --version 0.2 http://scruff.org/bzr/scruff.dev

    Note that the created tarball is just the same as the contents of
    the branch at the specified revision. If you wish to have something
    different, for instance the results of running "make dist", then you
    should create the tarball first, and pass it to the command as in
    the second example.
    """
    takes_args = ['location?', 'upstream_branch?']
    aliases = ['mu']

    package_opt = Option('package', help="The name of the source package.",
                         type=str)
    version_opt = Option('version',
        help="The upstream version number of this release, for example "
        "\"0.2\".", type=str)
    distribution_opt = Option('distribution', help="The distribution that "
            "this release is targetted at.", type=str)
    directory_opt = Option('directory',
                           help='Working tree into which to merge.',
                           short_name='d', type=unicode)
    last_version_opt = Option('last-version',
                              help='The full version of the last time '
                              'upstream was merged.', type=str)
    force_opt = Option('force',
                       help=('Force a merge even if the upstream branch '
                             'has not changed.'))
    snapshot_opt = Option('snapshot', help="Merge a snapshot from the "
            "upstream branch rather than a new upstream release.")

    launchpad_opt = Option('launchpad',
        help='Use Launchpad to find upstream locations.')

    takes_options = [package_opt, version_opt,
            distribution_opt, directory_opt, last_version_opt,
            force_opt, 'revision', 'merge-type',
            snapshot_opt, launchpad_opt]

    def _add_changelog_entry(self, tree, package, version, distribution_name,
            changelog):
        from bzrlib.plugins.builddeb.merge_upstream import (
            changelog_add_new_version)
        from bzrlib.plugins.builddeb.errors import (
            DchError,
            )
        try:
            changelog_add_new_version(tree, version, distribution_name,
                changelog, package)
        except DchError, e:
            note(e)
            raise BzrCommandError('Adding a new changelog stanza after the '
                    'merge had completed failed. Add the new changelog '
                    'entry yourself, review the merge, and then commit.')

    def _do_merge(self, tree, tarball_filenames, package, version,
            current_version, upstream_branch, upstream_revisions, merge_type,
            force):
        from bzrlib.plugins.builddeb.import_dsc import (
            DistributionBranch,
            DistributionBranchSet,
            )
        from bzrlib.plugins.builddeb.util import (
            component_from_orig_tarball,
            )
        db = DistributionBranch(tree.branch, tree.branch, tree=tree)
        dbs = DistributionBranchSet()
        dbs.add_branch(db)
        tarballs = [(p, component_from_orig_tarball(p, package, version)) for p
                in tarball_filenames]
        conflicts = db.merge_upstream(tarballs, package, version,
                current_version, upstream_branch=upstream_branch,
                upstream_revisions=upstream_revisions,
                merge_type=merge_type, force=force)
        return conflicts

    def _fetch_tarball(self, package, version, orig_dir, locations, v3):
        from bzrlib.plugins.builddeb.repack_tarball import repack_tarball
        from bzrlib.plugins.builddeb.util import tarball_name
        ret = []
        format = None
        for location in locations:
            if v3:
                if location.endswith(".tar.bz2") or location.endswith(".tbz2"):
                    format = "bz2"
                elif location.endswith(".tar.xz"):
                    format = "xz"
            dest_name = tarball_name(package, version, None, format=format)
            tarball_filename = os.path.join(orig_dir, dest_name)
            try:
                repack_tarball(location, dest_name, target_dir=orig_dir)
            except FileExists:
                raise BzrCommandError("The target file %s already exists, and is either "
                                      "different to the new upstream tarball, or they "
                                      "are of different formats. Either delete the target "
                                      "file, or use it as the argument to import."
                                      % dest_name)
            ret.append(tarball_filename)
        return ret

    def _get_tarballs(self, config, tree, package, version, upstream_branch,
            upstream_revision, v3, locations):
        orig_dir = config.orig_dir or default_orig_dir
        orig_dir = os.path.join(tree.basedir, orig_dir)
        if not os.path.exists(orig_dir):
            os.makedirs(orig_dir)
        return self._fetch_tarball(package, version, orig_dir,
            locations, v3)

    def run(self, location=None, upstream_branch=None, version=None,
            distribution=None, package=None,
            directory=".", revision=None, merge_type=None,
            last_version=None, force=None, snapshot=False, launchpad=False):
        try:
            from debian.changelog import Version
        except ImportError:
            # Prior to 0.1.15 the debian module was called debian_bundle
            from debian_bundle.changelog import Version

        from bzrlib.plugins.builddeb.errors import PackageVersionNotPresent
        from bzrlib.plugins.builddeb.hooks import run_hook
        from bzrlib.plugins.builddeb.upstream import (
            TarfileSource,
            UScanSource,
            )
        from bzrlib.plugins.builddeb.upstream.branch import (
            UpstreamBranchSource,
            )
        from bzrlib.plugins.builddeb.util import (
            FORMAT_3_0_QUILT,
            FORMAT_3_0_NATIVE,
            tree_get_source_format,
            guess_build_type,
            tree_contains_upstream_source,
            )

        tree, _ = WorkingTree.open_containing(directory)
        tree.lock_write()
        try:
            # Check for uncommitted changes.
            if tree.changes_from(tree.basis_tree()).has_changed():
                raise BzrCommandError("There are uncommitted changes in the "
                        "working tree. You must commit before using this "
                        "command.")
            config = debuild_config(tree, tree)
            (current_version, package, distribution, distribution_name,
             changelog, top_level) = _get_changelog_info(tree, last_version,
                 package, distribution)
            if package is None:
                raise BzrCommandError("You did not specify --package, and "
                        "there is no changelog from which to determine the "
                        "package name, which is needed to know the name to "
                        "give the .orig.tar.gz. Please specify --package.")

            contains_upstream_source = tree_contains_upstream_source(tree)
            if changelog is None:
                changelog_version = None
            else:
                changelog_version = changelog.version
            build_type = config.build_type
            if build_type is None:
                build_type = guess_build_type(tree, changelog_version,
                    contains_upstream_source)
            need_upstream_tarball = (build_type != BUILD_TYPE_MERGE)
            if build_type == BUILD_TYPE_NATIVE:
                raise BzrCommandError(gettext("Merge upstream in native mode "
                        "is not supported."))

            if launchpad:
                from bzrlib.plugins.builddeb.launchpad import (
                    get_upstream_branch_url as lp_get_upstream_branch_url,
                    )
                upstream_branch = lp_get_upstream_branch_url(package,
                    distribution_name, distribution)
                note(gettext("Using upstream branch %s") % upstream_branch)

            if upstream_branch is not None:
                upstream_branch = Branch.open(upstream_branch)
            elif location is not None:
                try:
                    upstream_branch = Branch.open(location)
                except NotBranchError:
                    upstream_branch = None
            elif upstream_branch is None and config.upstream_branch is not None:
                upstream_branch = Branch.open(config.upstream_branch)
            else:
                upstream_branch = None

            if upstream_branch is not None:
                upstream_branch_source = UpstreamBranchSource(
                    upstream_branch, config=config)
            else:
                upstream_branch_source = None

            if location is not None:
                try:
                    primary_upstream_source = UpstreamBranchSource(
                        Branch.open(location), config=config)
                except NotBranchError:
                    primary_upstream_source = TarfileSource(location, version)
            else:
                if snapshot:
                    if upstream_branch_source is None:
                        raise BzrCommandError(gettext("--snapshot requires "
                            "an upstream branch source"))
                    primary_upstream_source = upstream_branch_source
                else:
                    primary_upstream_source = UScanSource(tree, top_level)

            if revision is not None:
                if upstream_branch is None:
                    raise BzrCommandError(gettext("--revision can only be "
                        "used with a valid upstream branch"))
                if len(revision) > 1:
                    raise BzrCommandError(gettext("merge-upstream takes "
                        "only a single --revision"))
                upstream_revspec = revision[0]
                upstream_revisions = { None: upstream_revspec.as_revision_id(
                    upstream_branch) }
            else:
                upstream_revisions = None

            if version is None and upstream_revisions is not None:
                # Look up the version from the upstream revision
                version = upstream_branch_source.get_version(package,
                    current_version, upstream_revisions[None])
            elif version is None and primary_upstream_source is not None:
                version = primary_upstream_source.get_latest_version(
                    package, current_version)
            if version is None:
                if upstream_branch_source is not None:
                    raise BzrCommandError(gettext("You must specify "
                        "the version number using --version or specify "
                        "--snapshot to merge a snapshot from the upstream "
                        "branch."))
                else:
                    raise BzrCommandError(gettext("You must specify the "
                        "version number using --version."))
            assert isinstance(version, str)
            note(gettext("Using version string %s.") % (version))
            # Look up the revision id from the version string
            if upstream_revisions is None and upstream_branch_source is not None:
                try:
                    upstream_revisions = upstream_branch_source.version_as_revisions(
                        package, version)
                except PackageVersionNotPresent:
                    raise BzrCommandError(
                        "Version %s can not be found in upstream branch %r. "
                        "Specify the revision manually using --revision or adjust "
                        "'export-upstream-revision' in the configuration." %
                        (version, upstream_branch_source))
            if need_upstream_tarball:
                target_dir = tempfile.mkdtemp()
                self.add_cleanup(shutil.rmtree, target_dir)
                try:
                    locations = primary_upstream_source.fetch_tarballs(
                        package, version, target_dir, components=[None])
                except PackageVersionNotPresent:
                    if upstream_revisions is not None:
                        locations = upstream_branch_source.fetch_tarballs(
                            package, version, target_dir, components=[None],
                            revisions=upstream_revisions)
                    else:
                        raise
                source_format = tree_get_source_format(tree)
                v3 = (source_format in [
                    FORMAT_3_0_QUILT, FORMAT_3_0_NATIVE])
                tarball_filenames = self._get_tarballs(config, tree, package,
                    version, upstream_branch, upstream_revisions, v3,
                    locations)
                conflicts = self._do_merge(tree, tarball_filenames, package,
                    version, current_version, upstream_branch, upstream_revisions,
                    merge_type, force)
            if (current_version is not None and
                Version(current_version) >= Version(version)):
                raise BzrCommandError(
                    gettext("Upstream version %s has already been merged.") %
                    version)
            if not tree.has_filename("debian"):
                tree.mkdir("debian")
            self._add_changelog_entry(tree, package, version,
                distribution_name, changelog)
            run_hook(tree, 'merge-upstream', config)
        finally:
            tree.unlock()
        if not need_upstream_tarball:
            note(gettext("An entry for the new upstream version has been "
                 "added to the changelog."))
        else:
            note(gettext("The new upstream version has been imported."))
            if conflicts:
                note(gettext("You should now resolve the conflicts, review "
                     "the changes, and then commit."))
            else:
                note(gettext("You should now review the changes and "
                             "then commit."))


class cmd_import_dsc(Command):
    """Import a series of source packages.

    Provide a number of source packages (.dsc files), and they will
    be imported to create a branch with history that reflects those
    packages.

    The first argument is the distribution that these source packages
    were uploaded to, one of "debian" or "ubuntu". It can also
    be the target distribution from the changelog, e.g. "unstable",
    which will be resolved to the correct distribution.

    You can also specify a file (possibly remote) that contains a
    list of source packages (.dsc files) to import using the --file
    option. Each line is taken to be a URI or path to import. The
    sources specified in the file are used in addition to those
    specified on the command line.

    If you have an existing branch containing packaging and you want to
    import a .dsc from an upload done from outside the version control
    system you can use this command.
    """

    takes_args = ['files*']

    filename_opt = Option('file', help="File containing URIs of source "
                          "packages to import.", type=str, short_name='F')

    takes_options = [filename_opt]

    def import_many(self, db, files_list, orig_target):
        from bzrlib.plugins.builddeb.import_dsc import (
            DscCache,
            DscComp,
            )
        from bzrlib.plugins.builddeb.util import (
            open_file_via_transport,
            )
        cache = DscCache()
        files_list.sort(cmp=DscComp(cache).cmp)
        if not os.path.exists(orig_target):
            os.makedirs(orig_target)
        for dscname in files_list:
            dsc = cache.get_dsc(dscname)
            def get_dsc_part(from_transport, filename):
                from_f = open_file_via_transport(filename, from_transport)
                contents = from_f.read()
                to_f = open(os.path.join(orig_target, filename), 'wb')
                try:
                    to_f.write(contents)
                finally:
                    to_f.close()
            base, filename = urlutils.split(dscname)
            from_transport = cache.get_transport(dscname)
            get_dsc_part(from_transport, filename)
            for file_details in dsc['files']:
                name = file_details['name']
                get_dsc_part(from_transport, name)
            db.import_package(os.path.join(orig_target, filename))

    def run(self, files_list, file=None):
        from bzrlib.plugins.builddeb.errors import (
            MissingChangelogError,
            )
        from bzrlib.plugins.builddeb.import_dsc import (
            DistributionBranch,
            DistributionBranchSet,
            )
        from bzrlib.plugins.builddeb.util import (
            find_changelog,
            open_file,
            )
        try:
            tree = WorkingTree.open_containing('.')[0]
        except NotBranchError:
            raise BzrCommandError(gettext(
                "There is no tree to import the packages in to"))
        tree.lock_write()
        try:
            if tree.changes_from(tree.basis_tree()).has_changed():
                raise BzrCommandError(gettext("There are uncommitted "
                        "changes in the working tree. You must commit "
                        "before using this command"))
            if files_list is None:
                files_list = []
            if file is not None:
                if isinstance(file, unicode):
                    file = file.encode('utf-8')
                sources_file = open_file(file)
                for line in sources_file:
                    line = line.strip()
                    if len(line) > 0:
                        files_list.append(line)
            if len(files_list) < 1:
                raise BzrCommandError(gettext("You must give the location of "
                    "at least one source package to install, or use the "
                    "--file option."))
            config = debuild_config(tree, tree)
            if config.build_type == BUILD_TYPE_MERGE:
                raise BzrCommandError(
                    gettext("import-dsc in merge mode is not yet supported."))
            orig_dir = config.orig_dir or default_orig_dir
            orig_target = os.path.join(tree.basedir, default_orig_dir)
            db = DistributionBranch(tree.branch, tree.branch, tree=tree)
            dbs = DistributionBranchSet()
            dbs.add_branch(db)
            try:
                (changelog, top_level) = find_changelog(tree, False)
                last_version = changelog.version
            except MissingChangelogError:
                last_version = None
            tempdir = tempfile.mkdtemp(dir=os.path.join(tree.basedir,
                        '..'))
            try:
                if last_version is not None:
                    if not db.pristine_upstream_source.has_version(
                            changelog.package, last_version.upstream_version):
                        raise BzrCommandError(gettext("Unable to find the tag "
                            "for the previous upstream version, %(version)s, in the "
                            "branch. Consider importing it via import-upstream."
                            "If it is already present in the branch please "
                            "make sure it is tagged as %(tag)r.") %
                            {"version": last_version,
                                "tag": db.pristine_upstream_source.tag_name(
                                        last_version.upstream_version)})
                    upstream_tips = db.pristine_upstream_source.version_as_revisions(
                        changelog.package, last_version.upstream_version)
                    db.extract_upstream_tree(upstream_tips, tempdir)
                else:
                    db._create_empty_upstream_tree(tempdir)
                self.import_many(db, files_list, orig_target)
            finally:
                shutil.rmtree(tempdir)
        finally:
            tree.unlock()


class cmd_import_upstream(Command):
    """Imports an upstream tarball.

    This will import an upstream tarball in to your branch, but not modify the
    working tree. Use merge-upstream if you wish to directly merge the new
    upstream version in to your tree.

    The imported revision can be accessed using the tag name that will be
    reported at the end of a successful operation. The revision will include
    the pristine-tar data that will allow other commands to recreate the
    tarball when needed.

    For instance::

        $ bzr import-upstream 1.2.3 ../package_1.2.3.orig.tar.gz

    If upstream is packaged in bzr, you should provide the upstream branch
    whose tip commit is the closest match to the tarball::

        $ bzr import-upstream 1.2.3 ../package_1.2.3.orig.tar.gz ../upstream

    After doing this, commands that assume there is an upstream tarball, like
    'bzr builddeb' will be able to recreate the one provided at import-upstream
    time, meaning that you don't need to distribute the tarball in addition to
    the branch.

    If you want to manually merge with the imported upstream, you can do::

        $ bzr merge . -r tag:upstream-1.2.3

    The imported revision will have file ids taken from your branch, the
    upstream branch, or previous tarball imports as necessary. In addition
    the parents of the new revision will be the previous upstream tarball
    import and the tip of the upstream branch if you supply one.
    """

    takes_options = ['revision']

    takes_args = ['version', 'location', 'upstream_branch?']

    def run(self, version, location, upstream_branch=None, revision=None):
        try:
            from debian.changelog import Version
        except ImportError:
            # Prior to 0.1.15 the debian module was called debian_bundle
            from debian_bundle.changelog import Version
        from bzrlib.plugins.builddeb.import_dsc import (
            DistributionBranch,
            DistributionBranchSet,
            )
        from bzrlib.plugins.builddeb.util import (
            md5sum_filename,
            )
        # TODO: search for similarity etc.
        version = version.encode('utf8')
        branch, _ = Branch.open_containing('.')
        if upstream_branch is None:
            upstream = None
        else:
            upstream = Branch.open(upstream_branch)
        branch.lock_write() # we will be adding a tag here.
        self.add_cleanup(branch.unlock)
        tempdir = tempfile.mkdtemp(
            dir=branch.bzrdir.root_transport.clone('..').local_abspath('.'))
        self.add_cleanup(shutil.rmtree, tempdir)
        db = DistributionBranch(branch, pristine_upstream_branch=branch)
        if db.pristine_upstream_source.has_version(None, version):
            raise BzrCommandError(gettext("Version %s is already present.") % version)
        tagged_versions = {}
        for tversion, tcomponents in db.pristine_upstream_source.iter_versions():
            tagged_versions[Version(tversion)] = tcomponents
        tag_order = sorted(tagged_versions.keys())
        if tag_order:
            base_revisions = tagged_versions[tag_order[-1]]
        else:
            base_revisions = {}
        if base_revisions:
            if upstream is not None:
                # See bug lp:309682
                for parent in base_revisions.values():
                    upstream.repository.fetch(branch.repository, parent)
            db.extract_upstream_tree(base_revisions, tempdir)
            parents = {}
            for name, base_revid in base_revisions.iteritems():
                parents[name] = [base_revid]
        else:
            parents = {}
            db._create_empty_upstream_tree(tempdir)
        tree = db.branch.basis_tree()
        tree.lock_read()
        dbs = DistributionBranchSet()
        dbs.add_branch(db)
        if revision is None:
            upstream_revid = None
        elif len(revision) == 1:
            upstream_revid = revision[0].in_history(upstream).rev_id
        else:
            raise BzrCommandError(gettext(
                'bzr import-upstream --revision takes exactly'
                ' one revision specifier.'))
        tarballs = [(location, None, md5sum_filename(location))]
        for (component, tag_name, revid) in db.import_upstream_tarballs(
                tarballs, None, version, parents, upstream_branch=upstream,
                upstream_revisions={ None: upstream_revid }):
            if component is None:
                self.outf.write(gettext(
                    'Imported %(location)s as tag:%(tag)s.\n') % {
                        "location": location, "tag": tag_name})
            else:
                self.outf.write(gettext(
                    'Imported %(location)s (%(component)s) as tag:%(tag)s.\n') % {
                    "location": location,
                    "component": component,
                    "tag": tag_name})


class cmd_builddeb_do(Command):
    """Run a command in an exported package, copying the result back.

    For a merge mode package the full source is not available, making some
    operations difficult. This command allows you to run any command in an
    exported source directory, copying the resulting debian/ directory back
    to your branch if the command is successful.

    For instance:

      bzr builddeb-do

    will run a shell in the unpacked source. Any changes you make in the
    ``debian/`` directory (and only those made in that directory) will be copied
    back to the branch. If you exit with a non-zero exit code (e.g. "exit 1"),
    then the changes will not be copied back.

    You can also specify single commands to be run, e.g.

      bzr builddeb-do "dpatch-edit-patch 01-fix-build"

    Note that only the first argument is used as the command, and so the above
    example had to be quoted.
    """

    takes_args = ['command*']
    aliases = ['bd-do']

    def run(self, command_list=None):
        import subprocess
        from bzrlib.plugins.builddeb.errors import (
            BuildFailedError,
            )
        from bzrlib.plugins.builddeb.source_distiller import (
            MergeModeDistiller,
            )
        from bzrlib.plugins.builddeb.builder import (
            DebBuild,
            )
        from bzrlib.plugins.builddeb.upstream import (
            AptSource,
            DebianRulesSource,
            UScanSource,
            UpstreamProvider,
            )
        from bzrlib.plugins.builddeb.upstream.pristinetar import (
            PristineTarSource,
            )
        from bzrlib.plugins.builddeb.hooks import run_hook
        from bzrlib.plugins.builddeb.util import (
            find_changelog,
            guess_build_type,
            tree_contains_upstream_source,
            )
        t = WorkingTree.open_containing('.')[0]
        self.add_cleanup(t.lock_read().unlock)
        config = debuild_config(t, t)
        (changelog, top_level) = find_changelog(t, False, max_blocks=2)

        contains_upstream_source = tree_contains_upstream_source(t)
        if changelog is None:
            changelog_version = None
        else:
            changelog_version = changelog.version
        build_type = config.build_type
        if build_type is None:
            build_type = guess_build_type(t, changelog_version,
                contains_upstream_source)

        if build_type != BUILD_TYPE_MERGE:
            raise BzrCommandError(gettext("This command only works for merge "
                "mode packages. See /usr/share/doc/bzr-builddeb"
                "/user_manual/merge.html for more information."))

        give_instruction = False
        if command_list is None:
            try:
                command_list = [os.environ['SHELL']]
            except KeyError:
                command_list = ["/bin/sh"]
            give_instruction = True
        build_dir = config.build_dir
        if build_dir is None:
            build_dir = default_build_dir
        orig_dir = config.orig_dir
        if orig_dir is None:
            orig_dir = default_orig_dir

        upstream_provider = UpstreamProvider(changelog.package,
                changelog.version.upstream_version, orig_dir,
                [PristineTarSource(t, t.branch),
                 AptSource(),
                 DebianRulesSource(t, top_level),
                 UScanSource(t, top_level) ])

        distiller = MergeModeDistiller(t, upstream_provider,
                top_level=top_level)

        build_source_dir = os.path.join(build_dir,
                changelog.package + "-" + changelog.version.upstream_version)

        command = " ".join(command_list)

        builder = DebBuild(distiller, build_source_dir, command)
        builder.prepare()
        run_hook(t, 'pre-export', config)
        builder.export()
        note(gettext('Running "%s" in the exported directory.') % (command))
        if give_instruction:
            note(gettext('If you want to cancel your changes then exit '
                 'with a non-zero exit code, e.g. run "exit 1".'))
        try:
            builder.build()
        except BuildFailedError:
            raise BzrCommandError(gettext('Not updating the working tree as '
                'the command failed.'))
        note(gettext("Copying debian/ back"))
        if top_level:
            destination = ''
        else:
            destination = 'debian/'
        destination = os.path.join(t.basedir, destination)
        source_debian = os.path.join(build_source_dir, 'debian')
        for filename in os.listdir(source_debian):
            proc = subprocess.Popen('cp -apf "%s" "%s"' % (
                 os.path.join(source_debian, filename), destination),
                 shell=True)
            proc.wait()
            if proc.returncode != 0:
                raise BzrCommandError(gettext('Copying back debian/ failed'))
        builder.clean()
        note(gettext('If any files were added or removed you should run '
                     '"bzr add" or "bzr rm" as appropriate.'))


class cmd_mark_uploaded(Command):
    """Mark that this branch has been uploaded, prior to pushing it.

    When a package has been uploaded we want to mark the revision
    that it was uploaded in. This command automates doing that
    by marking the current tip revision with the version indicated
    in debian/changelog.
    """
    force = Option('force', help="Mark the upload even if it is already "
            "marked.")

    takes_options = [merge_opt, force]

    def run(self, merge=None, force=None):
        from bzrlib.plugins.builddeb.import_dsc import (
            DistributionBranch,
            DistributionBranchSet,
            )
        from bzrlib.plugins.builddeb.util import (
            find_changelog,
            )
        t = WorkingTree.open_containing('.')[0]
        t.lock_write()
        try:
            if t.changes_from(t.basis_tree()).has_changed():
              raise BzrCommandError(gettext("There are uncommitted "
                      "changes in the working tree. You must commit "
                      "before using this command"))
            config = debuild_config(t, t)
            if merge is None:
                merge = (config.build_type == BUILD_TYPE_MERGE)
            (changelog, top_level) = find_changelog(t, merge)
            if changelog.distributions == 'UNRELEASED':
                if not force:
                    raise BzrCommandError(gettext("The changelog still targets "
                            "'UNRELEASED', so apparently hasn't been "
                            "uploaded."))
            db = DistributionBranch(t.branch, None)
            dbs = DistributionBranchSet()
            dbs.add_branch(db)
            if db.has_version(changelog.version):
                if not force:
                    raise BzrCommandError(gettext(
                        "This version has already been "
                        "marked uploaded. Use --force to force marking "
                        "this new version."))
            tag_name = db.tag_version(changelog.version)
            self.outf.write(gettext("Tag '%s' created.\n") % tag_name)
        finally:
            t.unlock()


class cmd_merge_package(Command):
    """Merges source packaging branch into target packaging branch.

    This will first check whether the upstream branches have diverged.

    If that's the case an attempt will be made to fix the upstream ancestry
    so that the user only needs to deal with packaging branch merge issues.

    In the opposite case a normal merge will be performed.

    As of bzr 2.5 and bzr-builddeb 2.8.1, this functionality is automatically
    provided as part of bzr merge.
    """
    takes_options = ['revision']
    takes_args = ['source']

    if bzrlib_version >= (2, 5):
        hidden = True

    def run(self, source, revision=None):
        from bzrlib import ui
        from bzrlib.merge import Merger
        from bzrlib.tag import _merge_tags_if_possible
        from bzrlib.plugins.builddeb.merge_package import fix_ancestry_as_needed
        if 'pre_merge' in Merger.hooks:
            ui.ui_factory.show_warning(
                "The merge-package command is deprecated. Use 'bzr merge' "
                "instead.")
        source_branch = None
        # Get the target branch.
        try:
            tree = WorkingTree.open_containing('.')[0]
        except (NotBranchError, NoWorkingTree):
            raise BzrCommandError(
                gettext("There is no tree to merge the source branch in to"))
        # Get the source branch.
        try:
            source_branch = Branch.open(source)
        except NotBranchError:
            raise BzrCommandError(gettext("Invalid source branch URL?"))

        if revision is None:
            revid = source_branch.last_revision()
        elif len(revision) == 1:
            revid = revision[0].in_history(source_branch).rev_id
        else:
            raise BzrCommandError(gettext('bzr merge-package --revision takes '
                'exactly one argument'))

        tree.lock_write()
        self.add_cleanup(tree.unlock)
        source_branch.lock_read()
        self.add_cleanup(source_branch.unlock)
        this_config = debuild_config(tree, tree)
        that_config = debuild_config(source_branch.basis_tree(),
                source_branch.basis_tree())
        if not (this_config.build_type == BUILD_TYPE_NATIVE or
                that_config.build_type == BUILD_TYPE_NATIVE):
            fix_ancestry_as_needed(tree, source_branch, source_revid=revid)

        # Merge source packaging branch in to the target packaging branch.
        _merge_tags_if_possible(source_branch, tree.branch)
        conflicts = tree.merge_from_branch(source_branch, to_revision=revid)
        if conflicts > 0:
            note(gettext(
                'The merge resulted in %s conflicts. Please resolve these '
                'and commit the changes with "bzr commit".') % conflicts)
        else:
            note(gettext('The merge resulted in no conflicts. You may '
                'commit the changes by running "bzr commit".'))


class cmd_dh_make(Command):
    """Helps you create a new package.

    This code wraps dh_make to do the Bazaar setup for you, ensuring that
    your branches have all the necessary information and are correctly
    linked to the upstream branches where necessary.

    The basic use case is satisfied by

        bzr dh-make project 0.1 http://project.org/project-0.1.tar.gz

    which will import the tarball with the correct tags etc. and then
    run dh_make for you in order to start the packaging.

    If there upstream is available in bzr then run the command from the
    root of a branch of that corresponding to the 0.1 release.

    If there is no upstream available in bzr then run the command from
    outside a branch and it will create a branch for you in a directory
    named the same as the package name you specify as the second argument.

    If you do not wish to use dh_make, but just take advantage of the
    Bazaar specific parts then use the --bzr-only option.
    """

    aliases = ['dh_make']

    takes_args = ['package_name', 'version', 'tarball']

    bzr_only_opt = Option('bzr-only', help="Don't run dh_make.")
    v3_opt = Option('v3', hidden=True)

    takes_options = [bzr_only_opt, v3_opt]

    def run(self, package_name, version, tarball, bzr_only=None, v3=None):
        from bzrlib.plugins.builddeb import dh_make
        if v3:
            warning(gettext('The --v3 option has been deprecated. dh_make now '
                            'creates format v3 packages by default.'))
        tree = dh_make.import_upstream(tarball, package_name,
            version.encode("utf-8"))
        if not bzr_only:
            tree.lock_write()
            try:
                dh_make.run_dh_make(tree, package_name, version)
            finally:
                tree.unlock()
        note(gettext('Package prepared in %s'),
            urlutils.unescape_for_display(tree.basedir, self.outf.encoding))


class cmd_dep3_patch(Command):
    """Format the changes in a branch as a DEP-3 patch.

    This will generate a patch file containing as much information
    specified by DEP-3 (http://dep.debian.net/deps/dep3/) as possible.

    The patch will contain all changes that are not merged into
    the current branch (either that in the current working directory
    or specified by --directory) but are present and committed
    in the branch at the specified location.

    To generate the "Status" header, this command will check the
    upstream branch to verify if the change has made it upstream,
    unless --no-upstream-check was specified.

    examples::

        bzr dep3-patch lp:~user/project/feature-branch > \\
                debian/patches/01_feature
    """

    takes_args = ["location?"]

    directory_opt = Option('directory',
                           help='Packaging tree for which to generate patch.',
                           short_name='d', type=unicode)

    no_upstream_check_opt = Option('no-upstream-check',
        help="Don't check whether patch has been merged upstream.")

    takes_options = [directory_opt, "revision", "change", no_upstream_check_opt]

    def run(self, location=".", directory=".", revision=None,
            no_upstream_check=False):
        from bzrlib.plugins.builddeb.dep3 import (
            determine_applied_upstream,
            determine_forwarded,
            describe_origin,
            gather_bugs_and_authors,
            write_dep3_patch,
            )
        packaging_tree, packaging_branch = BzrDir.open_containing_tree_or_branch(
            directory)[:2]
        self.add_cleanup(packaging_branch.lock_read().unlock)
        tree, branch = BzrDir.open_containing_tree_or_branch(location)[:2]
        self.add_cleanup(branch.lock_read().unlock)
        if revision is not None and len(revision) >= 1:
            revision_id = revision[-1].as_revision_id(branch)
        else:
            revision_id = None
        if revision_id is None:
            revision_id = branch.last_revision()
        graph = branch.repository.get_graph(packaging_branch.repository)
        if revision is not None and len(revision) == 2:
            base_revid = revision[0].as_revision_id(branch)
        else:
            base_revid = graph.find_unique_lca(revision_id,
                packaging_branch.last_revision())
        interesting_revision_ids = graph.find_unique_ancestors(revision_id,
            [base_revid])
        if len(interesting_revision_ids) == 0:
            raise BzrCommandError(gettext("No unmerged revisions"))
        (bugs, authors, last_update) = gather_bugs_and_authors(
            branch.repository, interesting_revision_ids)
        config = branch.get_config()
        description = config.get_user_option("description")
        if description is None:
            # if there's just one revision in the mainline history, use
            # that revisions commits message
            lhs_history = graph.iter_lefthand_ancestry(revision_id,
                [base_revid])
            rev = branch.repository.get_revision(lhs_history.next())
            try:
                lhs_history.next()
            except StopIteration:
                description = rev.message
        origin = describe_origin(branch, revision_id)
        if packaging_tree is None:
            packaging_tree = packaging_branch.basis_tree()
        builddeb_config = debuild_config(packaging_tree, True)
        if not no_upstream_check and builddeb_config.upstream_branch:
            upstream_branch = Branch.open(builddeb_config.upstream_branch)
            applied_upstream = determine_applied_upstream(upstream_branch,
                branch, revision_id)
            forwarded = determine_forwarded(upstream_branch, branch,
                revision_id)
        else:
            applied_upstream = None
            forwarded = None
        write_dep3_patch(self.outf, branch, base_revid,
            revision_id, bugs=bugs, authors=authors, origin=origin,
            forwarded=forwarded, applied_upstream=applied_upstream,
            description=description, last_update=last_update)



#    __init__.py -- The plugin for bzr
#    Copyright (C) 2005 Jamie Wilkinson <jaq@debian.org> 
#                  2006, 2007 James Westby <jw+debian@jameswestby.net>
#                  2007 Reinhard Tartler <siretart@tauware.de>
#                  2008 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""bzr-builddeb - manage packages in a Bazaar branch."""

from __future__ import absolute_import

import os

import bzrlib
from bzrlib.commands import plugin_cmds

from bzrlib.plugins.builddeb.info import (
    bzr_plugin_version as version_info,
    )


try:
    from bzrlib.i18n import load_plugin_translations
except ImportError: # No translations for bzr < 2.5
    gettext = lambda x: x
else:
    translation = load_plugin_translations("bzr-builddeb")
    gettext = translation.ugettext

commands = {
        "builddeb_do": ["bd-do"],
        "builddeb": ["bd"],
        "get_orig_source": [],
        "dep3_patch": [],
        "dh_make": ["dh_make"],
        "import_dsc": [],
        "import_upstream": [],
        "mark_uploaded": [],
        "merge_package": [],
        "merge_upstream": ["mu"],
        }

for command, aliases in commands.iteritems():
    plugin_cmds.register_lazy('cmd_' + command, aliases, 
        "bzrlib.plugins.builddeb.cmds")

builddeb_dir = '.bzr-builddeb'
default_conf = os.path.join(builddeb_dir, 'default.conf')
def global_conf():
    from bzrlib.config import config_dir
    return os.path.join(config_dir(), 'builddeb.conf')
local_conf = os.path.join(builddeb_dir, 'local.conf')
new_local_conf = 'debian/local.conf.local'
new_conf = 'debian/bzr-builddeb.conf'

default_build_dir = '../build-area'
default_orig_dir = '..'
default_result_dir = '..'


try:
    from bzrlib.registry import register_lazy
except ImportError:
    from bzrlib.directory_service import (
        AliasDirectory,
        directories,
        )

    directories.register_lazy("apt:", 'bzrlib.plugins.builddeb.directory',
            'VcsDirectory',
            "Directory that uses Debian Vcs-* control fields to look up branches")

    branch_aliases = getattr(AliasDirectory, "branch_aliases", None)
    if branch_aliases is not None:
        branch_aliases.register_lazy("upstream",
                "bzrlib.plugins.builddeb.directory", "upstream_branch_alias",
                help="upstream branch (for packaging branches)")

    try:
        from bzrlib.tag import tag_sort_methods
    except ImportError:
        pass # bzr tags --sort= can not be extended
    else:
        tag_sort_methods.register_lazy("debversion",
            "bzrlib.plugins.builddeb.tagging", "sort_debversion",
            "Sort like Debian versions.")

    try:
        from bzrlib.revisionspec import revspec_registry
        revspec_registry.register_lazy("package:",
            "bzrlib.plugins.builddeb.revspec", "RevisionSpec_package")
        revspec_registry.register_lazy("upstream:",
            "bzrlib.plugins.builddeb.revspec", "RevisionSpec_upstream")
    except ImportError:
        from bzrlib.revisionspec import SPEC_TYPES
        from bzrlib.plugins.builddeb.revspec import (
            RevisionSpec_package,
            RevisionSpec_upstream,
            )
        SPEC_TYPES.extend([RevisionSpec_package, RevisionSpec_upstream])
else:
    register_lazy("bzrlib.directory", "directories", "apt:",
            'bzrlib.plugins.builddeb.directory', 'VcsDirectory',
            help="Directory that uses Debian Vcs-* control fields to look up branches")
    register_lazy("bzrlib.directory", "AliasDirectory.branch_aliases", "upstream",
            "bzrlib.plugins.builddeb.directory", "upstream_branch_alias",
            help="upstream branch (for packaging branches)")
    register_lazy("bzrlib.tag", "tag_sort_methods", "debversion",
        "bzrlib.plugins.builddeb.tagging", "sort_debversion",
        "Sort like Debian versions.")
    register_lazy("bzrlib.revisionspec", "revspec_registry", "package:",
        "bzrlib.plugins.builddeb.revspec", "RevisionSpec_package")
    register_lazy("bzrlib.revisionspec", "revspec_registry", "upstream:",
        "bzrlib.plugins.builddeb.revspec", "RevisionSpec_upstream")


def debian_changelog_commit_message(commit, start_message):
    if start_message is not None:
        return start_message
    cl_path = "debian/changelog"
    if not commit.work_tree.has_filename(cl_path):
        return start_message
    if commit.work_tree.path2id(cl_path) is None:
        return start_message
    if cl_path in commit.exclude:
        return start_message
    if commit.specific_files and cl_path not in commit.specific_files:
        return start_message
    changes = []
    for change in commit.work_tree.iter_changes(commit.work_tree.basis_tree(),
            specific_files=[cl_path]):
        # Content not changed
        if not change[2]:
            return start_message
        # Not versioned in new tree
        if not change[3][1]:
            return start_message
        # Not a file in one tree
        if change[6][0] != 'file' or change[6][1] != 'file':
            return start_message
        old_text = commit.work_tree.basis_tree().get_file(change[0],
                path=change[1][0]).readlines()
        new_text = commit.work_tree.get_file(change[0],
                path=change[1][1]).readlines()
        import difflib
        sequencematcher = difflib.SequenceMatcher
        for group in sequencematcher(None, old_text,
                new_text).get_grouped_opcodes(0):
            j1, j2 = group[0][3], group[-1][4]
            for line in new_text[j1:j2]:
                if line.startswith("  "):
                    changes.append(line)
    if not changes:
        return start_message
    from bzrlib.plugins.builddeb.util import strip_changelog_message
    changes = strip_changelog_message(changes)
    return "".join(changes)


def debian_changelog_commit(commit, start_message):
    """hooked into bzrlib.msgeditor set_commit_message.
     Set the commit message from debian/changelog and set any LP: #1234 to bug
     fixed tags."""
    from bzrlib.plugins.builddeb.util import (
        debuild_config, find_bugs_fixed)

    t = commit.work_tree
    config = debuild_config(t, False)
    if config.commit_message_from_changelog == False:
        return None

    changes = debian_changelog_commit_message(commit, start_message)
    if changes is None:
        return None

    bugs_fixed = find_bugs_fixed([changes], commit.work_tree.branch)
    commit.builder._revprops["bugs"] = "\n".join(bugs_fixed)

    # Debian Policy Manual states that debian/changelog must be UTF-8
    return changes.decode("utf-8")


def changelog_merge_hook_factory(merger):
    from bzrlib.plugins.builddeb import merge_changelog
    return merge_changelog.ChangeLogFileMerge(merger)


def debian_tag_name(branch, revid):
    from bzrlib.plugins.builddeb.config import BUILD_TYPE_MERGE
    from bzrlib.plugins.builddeb.errors import MissingChangelogError
    from bzrlib.plugins.builddeb.import_dsc import (DistributionBranch,
        DistributionBranchSet)
    from bzrlib.plugins.builddeb.util import debuild_config, find_changelog
    t = branch.repository.revision_tree(revid)
    config = debuild_config(t, False)
    try:
        (changelog, top_level) = find_changelog(t, config.build_type == BUILD_TYPE_MERGE)
    except MissingChangelogError:
        # Not a debian package
        return None
    if changelog.distributions == 'UNRELEASED':
        # The changelog still targets 'UNRELEASED', so apparently hasn't been 
        # uploaded. XXX: Give a warning of some sort here?
        return None
    db = DistributionBranch(branch, None)
    dbs = DistributionBranchSet()
    dbs.add_branch(db)
    return db.tag_name(changelog.version)


def start_commit_check_quilt(tree):
    """start_commit hook which checks the state of quilt patches.
    """
    if tree.path2id("debian/patches") is None:
        # No patches to worry about
        return
    from bzrlib.plugins.builddeb.merge_quilt import start_commit_quilt_patches
    start_commit_quilt_patches(tree)


def pre_merge(merger):
    pre_merge_fix_ancestry(merger)
    pre_merge_quilt(merger)


def pre_merge_quilt(merger):
    if getattr(merger, "_no_quilt_unapplying", False):
        return

    if (merger.other_tree.path2id("debian/patches/series") is None and
        merger.this_tree.path2id("debian/patches/series") is None and
        merger.working_tree.path2id("debian/patches/series") is None):
        return

    from bzrlib import trace
    from bzrlib.plugins.builddeb.util import debuild_config
    config = debuild_config(merger.working_tree, merger.working_tree)
    merger.debuild_config = config
    if not config.quilt_smart_merge:
        trace.mutter("skipping smart quilt merge, not enabled.")
        return

    if (merger.other_tree.path2id(".pc/applied-patches") is None and
        merger.this_tree.path2id(".pc/applied-patches") is None and
        merger.working_tree.path2id(".pc/applied-patches") is None):
        return

    from bzrlib.plugins.builddeb.errors import QuiltUnapplyError
    from bzrlib.plugins.builddeb.quilt import quilt_pop_all, quilt_series, QuiltError
    from bzrlib.plugins.builddeb.merge_quilt import tree_unapply_patches
    trace.note("Unapplying quilt patches to prevent spurious conflicts")
    merger._quilt_tempdirs = []
    merger._old_quilt_series = quilt_series(merger.working_tree)
    if merger._old_quilt_series:
        quilt_pop_all(working_dir=merger.working_tree.basedir)
    try:
        merger.this_tree, this_dir = tree_unapply_patches(merger.this_tree,
            merger.this_branch, force=True)
    except QuiltError, e:
        raise QuiltUnapplyError("this", e.stderr)
    else:
        if this_dir is not None:
            merger._quilt_tempdirs.append(this_dir)
    try:
        merger.base_tree, base_dir = tree_unapply_patches(merger.base_tree,
            merger.this_branch, force=True)
    except QuiltError, e:
        raise QuiltUnapplyError("base", e.stderr)
    else:
        if base_dir is not None:
            merger._quilt_tempdirs.append(base_dir)
    other_branch = getattr(merger, "other_branch", None)
    if other_branch is None:
        other_branch = merger.this_branch
    try:
        merger.other_tree, other_dir = tree_unapply_patches(merger.other_tree,
            other_branch, force=True)
    except QuiltError, e:
        raise QuiltUnapplyError("other", e.stderr)
    else:
        if other_dir is not None:
            merger._quilt_tempdirs.append(other_dir)


def post_merge_quilt_cleanup(merger):
    import shutil
    for dir in getattr(merger, "_quilt_tempdirs", []):
        shutil.rmtree(dir)
    config = getattr(merger, "debuild_config", None)
    if config is None:
        # If there is no debuild config, then pre_merge didn't get far enough.
        return
    policy = config.quilt_tree_policy
    if policy is None:
        return
    from bzrlib.plugins.builddeb.merge_quilt import post_process_quilt_patches
    post_process_quilt_patches(
        merger.working_tree,
        getattr(merger, "_old_quilt_series", []), policy)


def post_build_tree_quilt(tree):
    from bzrlib.plugins.builddeb.util import debuild_config
    config = debuild_config(tree, tree)
    policy = config.quilt_tree_policy
    if policy is None:
        return
    from bzrlib.plugins.builddeb.merge_quilt import post_process_quilt_patches
    post_process_quilt_patches(tree, [], policy)


def pre_merge_fix_ancestry(merger):
    from bzrlib.plugins.builddeb.config import BUILD_TYPE_NATIVE
    from bzrlib.plugins.builddeb.util import debuild_config
    from bzrlib.plugins.builddeb.merge_package import fix_ancestry_as_needed
    from bzrlib.workingtree import WorkingTree
    if not isinstance(merger.this_tree, WorkingTree):
        return
    if getattr(merger, "other_branch", None) is None:
        return
    if (not merger.this_tree.path2id("debian/changelog") or
        not merger.other_tree.path2id("debian/changelog")):
        return
    this_config = debuild_config(merger.this_tree, merger.this_tree)
    other_config = debuild_config(merger.other_tree, merger.other_tree)
    if not (this_config.build_type == BUILD_TYPE_NATIVE or
            other_config.build_type == BUILD_TYPE_NATIVE):
        from bzrlib import trace
        from bzrlib.plugins.builddeb.errors import PackageVersionNotPresent
        try:
            fix_ancestry_as_needed(merger.this_tree, merger.other_branch,
                source_revid=merger.other_tree.get_revision_id())
        except PackageVersionNotPresent, e:
            trace.warning(
                gettext("Not attempting to fix packaging branch ancestry, missing pristine tar "
                "data for version %s."), e.version)


try:
    from bzrlib.hooks import install_lazy_named_hook
except ImportError: # Compatibility with bzr < 2.4
    from bzrlib import (
        branch as _mod_branch,
        errors,
        merge,
        msgeditor,
        )
    msgeditor.hooks.install_named_hook("commit_message_template",
            debian_changelog_commit_message,
            "Use changes documented in debian/changelog to suggest "
            "the commit message")
    if getattr(merge, 'ConfigurableFileMerger', None) is None:
        raise ImportError(
            'need at least bzr 2.1.0rc2 (you use %r)', bzrlib.version_info)
    else:
        merge.Merger.hooks.install_named_hook(
            'merge_file_content', changelog_merge_hook_factory,
            'Debian Changelog file merge')
    try:
        _mod_branch.Branch.hooks.install_named_hook("automatic_tag_name",
             debian_tag_name,
             "Automatically determine tag names from Debian version")
    except errors.UnknownHook:
        pass # bzr < 2.2 doesn't have this hook.
else:
    install_lazy_named_hook(
        "bzrlib.msgeditor", "hooks", "commit_message_template",
            debian_changelog_commit_message,
            "Use changes documented in debian/changelog to suggest "
            "the commit message")
    if bzrlib.version_info[0] >= 2 and bzrlib.version_info[1] >= 4:
        install_lazy_named_hook(
            "bzrlib.msgeditor", "hooks", "set_commit_message",
                debian_changelog_commit,
                "Use changes documented in debian/changelog to set "
                "the commit message and bugs fixed")
    install_lazy_named_hook(
        "bzrlib.merge", "Merger.hooks",
        'merge_file_content', changelog_merge_hook_factory,
        'Debian Changelog file merge')
    install_lazy_named_hook(
        "bzrlib.branch", "Branch.hooks",
        "automatic_tag_name", debian_tag_name,
         "Automatically determine tag names from Debian version")
    install_lazy_named_hook(
        "bzrlib.merge", "Merger.hooks",
        'pre_merge', pre_merge,
        'Debian quilt patch (un)applying and ancestry fixing')
    install_lazy_named_hook(
        "bzrlib.merge", "Merger.hooks",
        'post_merge', post_merge_quilt_cleanup,
        'Cleaning up quilt temporary directories')
    install_lazy_named_hook(
        "bzrlib.mutabletree", "MutableTree.hooks",
        'post_build_tree', post_build_tree_quilt,
        'Applying quilt patches.')
    install_lazy_named_hook(
        "bzrlib.mutabletree", "MutableTree.hooks",
        "start_commit", start_commit_check_quilt,
        "Check for (un)applied quilt patches")

def load_tests(standard_tests, module, loader):
    return loader.loadTestsFromModuleNames(['bzrlib.plugins.builddeb.tests'])
